# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

from .defaults import init_defaults

from blinker import Signal
from collections import abc


class Settings(abc.Mapping):
    """Specialized mapping for *Value instances"""

    def __init__(self):
        self._defaults = {}
        self._values = {}
        self._constructors = {}
        self._descriptions = {}
        self._signals = {}
        self._global_signal = Signal()

    def add(self, name, constructor, default, description=None):
        """
        Add new setting

        name:        Identifier for this setting
        constructor: Callable that takes one argument and returns a new value
                     for this setting
        default:     Initial and default value
        description: What the setting does
        """
        self._constructors[name] = constructor
        self._signals[name] = Signal()
        self._descriptions[name] = description
        self[name] = default
        self._defaults[name] = self[name]
        self._global_signal.send(self, name=name, value=self[name])

    def reset(self, name):
        """Reset setting `name` to default/initial value"""
        self[name] = self._defaults[name]

    def default(self, name):
        """Return settings default/initial value"""
        return self._defaults[name]

    def description(self, name):
        """Return setting's description"""
        return self._descriptions[name]

    def syntax(self, name):
        """Return setting's description"""
        return self._constructors[name].syntax

    def validate(self, name, value):
        """Pass `value` to `name`'s constructor and return the result"""
        if not isinstance(value, str) and isinstance(value, abc.Iterable):
            return self._constructors[name](*value)
        else:
            return self._constructors[name](value)

    def on_change(self, callback, name=None, autoremove=True):
        """
        Run `callback` every time a value changes

        If `name` is None, run `callback` if any signal changes.

        The signature of `callback` must be: (settings, name, value)

        If `autoremove` is True, stop calling `callback` once it is garbage
        collected.
        """
        if name is None:
            self._global_signal.connect(callback, weak=autoremove)
        else:
            self._signals[name].connect(callback, weak=autoremove)

    def __getitem__(self, name):
        return self._values[name]

    def __setitem__(self, name, value):
        value_ = self.validate(name, value)
        self._values[name] = value_
        self._global_signal.send(self, name=name, value=value_)
        self._signals[name].send(self, name=name, value=value_)

    def __contains__(self, name):
        return name in self._constructors

    def __iter__(self):
        return iter(self._constructors)

    def __len__(self):
        return len(self._constructors)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import textwrap

class ClientError(Exception):
    def __init__(self, message):
        super().__init__(str(message))

    # Making exceptions with the same arguments equal helps in the tests
    def __eq__(self, other):
        return type(self) == type(other) and self.args == other.args

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(type(self).__name__ + str(self))

class ConnectionError(ClientError):
    def __init__(self, url):
        super().__init__('Failed to connect: %s' % url)

class TimeoutError(ClientError):
    def __init__(self, timeout, url):
        super().__init__('Timeout after %ss: %s' % (timeout, url))

class RPCError(ClientError):
    def __init__(self, response):
        super().__init__('Invalid RPC response: %s' %
                         textwrap.shorten(response, 100, placeholder='...'))

class AuthError(ClientError):
    def __init__(self, url):
        super().__init__('Authentication failed: %s' % url)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Monkey patches that should be removed when they are resolved upstream"""

import urwid

# Add more actions for key bindings
urwid.CURSOR_WORD_LEFT         = 'cursor word left'
urwid.CURSOR_WORD_RIGHT        = 'cursor word right'
urwid.DELETE_TO_EOL            = 'delete to end of line'
urwid.DELETE_LINE              = 'delete line'
urwid.DELETE_CHAR_UNDER_CURSOR = 'delete char under cursor'
urwid.DELETE_WORD_LEFT         = 'delete word left'
urwid.DELETE_WORD_RIGHT        = 'delete word right'
urwid.CANCEL                   = 'cancel'

# Remove urwid's default keybindings and create our own built-in command_map
for key in tuple(urwid.command_map._command):
    del urwid.command_map._command[key]

from .keymap import Key
urwid.command_map[Key('pgup')]           = urwid.CURSOR_PAGE_UP
urwid.command_map[Key('pgdn')]           = urwid.CURSOR_PAGE_DOWN
urwid.command_map[Key('ctrl-b')]         = urwid.CURSOR_PAGE_UP
urwid.command_map[Key('ctrl-f')]         = urwid.CURSOR_PAGE_DOWN
urwid.command_map[Key('b')]              = urwid.CURSOR_PAGE_UP
urwid.command_map[Key('space')]          = urwid.CURSOR_PAGE_DOWN

urwid.command_map[Key('up')]             = urwid.CURSOR_UP
urwid.command_map[Key('down')]           = urwid.CURSOR_DOWN
urwid.command_map[Key('left')]           = urwid.CURSOR_LEFT
urwid.command_map[Key('right')]          = urwid.CURSOR_RIGHT
urwid.command_map[Key('meta-b')]         = urwid.CURSOR_WORD_LEFT
urwid.command_map[Key('meta-f')]         = urwid.CURSOR_WORD_RIGHT

urwid.command_map[Key('home')]           = urwid.CURSOR_MAX_LEFT
urwid.command_map[Key('end')]            = urwid.CURSOR_MAX_RIGHT
urwid.command_map[Key('ctrl-a')]         = urwid.CURSOR_MAX_LEFT
urwid.command_map[Key('ctrl-e')]         = urwid.CURSOR_MAX_RIGHT

urwid.command_map[Key('ctrl-k')]         = urwid.DELETE_TO_EOL
urwid.command_map[Key('ctrl-u')]         = urwid.DELETE_LINE
urwid.command_map[Key('ctrl-d')]         = urwid.DELETE_CHAR_UNDER_CURSOR
urwid.command_map[Key('meta-d')]         = urwid.DELETE_WORD_LEFT
urwid.command_map[Key('meta-backspace')] = urwid.DELETE_WORD_RIGHT
urwid.command_map[Key('ctrl-w')]         = urwid.DELETE_WORD_RIGHT

urwid.command_map[Key('enter')]          = urwid.ACTIVATE
urwid.command_map[Key('escape')]         = urwid.CANCEL
urwid.command_map[Key('ctrl-g')]         = urwid.CANCEL
urwid.command_map[Key('ctrl-c')]         = urwid.CANCEL
urwid.command_map[Key('ctrl-l')]         = urwid.REDRAW_SCREEN


import re
import operator
class Edit_readline(urwid.Edit):
    def keypress(self, size, key):
        def move_to_next_word(forward=True):
            if forward:
                match_iterator  = re.finditer(r'(\b\W+|$)', self.edit_text, flags=re.UNICODE)
                match_positions = (m.start() for m in match_iterator)
                op = operator.gt
            else:
                match_iterator  = re.finditer(r'(\w+\b|^)', self.edit_text, flags=re.UNICODE)
                match_positions = reversed([m.start() for m in match_iterator])
                op = operator.lt
            for pos in match_positions:
                if op(pos, self.edit_pos):
                    self.set_edit_pos(pos)
                    return pos

        cmd = self._command_map[key]
        if cmd is urwid.DELETE_TO_EOL:
            self.edit_text = self.edit_text[:self.edit_pos]
            return None
        elif cmd is urwid.DELETE_LINE:
            self.set_edit_text('')
            return None
        elif cmd is urwid.DELETE_CHAR_UNDER_CURSOR:
            return super().keypress(size, 'delete')
        elif cmd is urwid.CURSOR_WORD_RIGHT:
            move_to_next_word(forward=True)
            return None
        elif cmd is urwid.CURSOR_WORD_LEFT:
            move_to_next_word(forward=False)
            return None
        elif cmd is urwid.DELETE_WORD_LEFT:
            start_pos = self.edit_pos
            end_pos = move_to_next_word(forward=True)
            if end_pos is not None:
                self.set_edit_text(self.edit_text[:start_pos] + self.edit_text[end_pos:])
            self.edit_pos = start_pos
            return None
        elif cmd is urwid.DELETE_WORD_RIGHT:
            end_pos = self.edit_pos
            start_pos = move_to_next_word(forward=False)
            if start_pos is not None:
                self.set_edit_text(self.edit_text[:start_pos] + self.edit_text[end_pos:])
            return None
        elif key == 'space':
            return super().keypress(size, ' ')
        else:
            return super().keypress(size, key)

urwid.Edit = Edit_readline


# Remove Text._calc_line_translation() to fix rare crash on peer lists.
# https://github.com/urwid/urwid/pull/307
#
# This bug seems impossible, but I've added extensive debugging messages and the
# issue is as follows:
#
# >>> def _update_cache_translation(self, maxcol, ta):
# >>>     self._cache_maxcol = maxcol  # maxcol is a number
# >>>     self._calc_line_translation(maxcol)
# >>>
# >>> def _calc_line_translation(self, maxcol):
# >>>     # self._cache_maxcol is now None while maxcol is still a number.

class Text_patched(urwid.Text):
    def _update_cache_translation(self, maxcol, ta):
        if ta:
            text, attr = ta
        else:
            text, attr = self.get_text()
        self._cache_maxcol = maxcol
        self._cache_translation = self.layout.layout(
            text, maxcol, self._align_mode, self._wrap_mode)

urwid.Text = Text_patched


# Limit the impact of the high CPU load bug
# https://github.com/urwid/urwid/pull/86
# https://github.com/urwid/urwid/issues/90
urwid.AsyncioEventLoop._idle_emulation_delay = 1/25


# urwid uses sys.exc_info() to get `(type, value, traceback)`, which is needed
# in python2 to re-raise exceptions.  However, sys.exc_info() can return `(None,
# None, None)` which leaves us with nothing to re-raise. It's possible this
# happens only with the asyncio event loop.
#
# This is easily solved for asyncio/python3 since we only need the exception
# which is guaranteed by asyncio.
class AsyncioEventLoop_patched(urwid.AsyncioEventLoop):
    def _exception_handler(self, loop, context):
        exc = context.get('exception')
        if exc:
            loop.stop()
            if not isinstance(exc, urwid.ExitMainLoop):
                self._exc_info = exc
        else:
            loop.default_exception_handler(context)

    def run(self):
        self._loop.set_exception_handler(self._exception_handler)
        self._loop.run_forever()
        if self._exc_info:
            raise self._exc_info

urwid.AsyncioEventLoop = AsyncioEventLoop_patched


class ListBox_patched(urwid.ListBox):
    # Add support for ScrollBar class (see stig.tui.scroll)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._rows_max = None

    def _invalidate(self):
        super()._invalidate()
        self._rows_max = None

    def get_scrollpos(self, size, focus=False):
        """Current scrolling position

        Lower limit is 0, upper limit is the highest index of `body`.
        """
        middle, top, bottom = self.calculate_visible(size, focus)
        if middle is None:
            return 0
        else:
            offset_rows, _, focus_pos, _, _ = middle
            maxcol, maxrow = size
            flow_size = (maxcol,)

            body = self.body
            if hasattr(body, 'positions'):
                # For body[pos], pos can be anything, not just an int.  In that
                # case, the positions() method returns an interable of valid
                # positions.
                positions = tuple(self.body.positions())
                focus_index = positions.index(focus_pos)
                widgets_above_focus = (body[pos] for pos in positions[:focus_index])
            else:
                # Treat body like a normal list
                widgets_above_focus = (w for w in body[:focus_pos])

            rows_above_focus = sum(w.rows(flow_size) for w in widgets_above_focus)
            rows_above_top = rows_above_focus - offset_rows
            return rows_above_top

    def rows_max(self, size, focus=False):
        if self._rows_max is None:
            flow_size = (size[0],)
            body = self.body
            if hasattr(body, 'positions'):
                self._rows_max = sum(body[pos].rows(flow_size) for pos in body.positions())
            else:
                self._rows_max = sum(w.rows(flow_size) for w in self.body)
        return self._rows_max

urwid.ListBox = ListBox_patched



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from .. import (InitCommand, CmdError, ExpectedResource)
from . import _mixin as mixin
from ._common import (make_X_FILTER_spec, make_COLUMNS_doc,
                      make_SORT_ORDERS_doc, make_SCRIPTING_doc)

import asyncio


class ListTorrentsCmdbase(mixin.get_torrent_sorter, mixin.get_torrent_columns,
                          metaclass=InitCommand):
    name = 'list'
    aliases = ('ls',)
    provides = set()
    category = 'torrent'
    description = 'List torrents'
    usage = ('list [<OPTIONS>]',
             'list [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('ls active',
                'ls !active',
                'ls seeds<10',
                'ls active&tracker~example.org',
                'ls active|idle&tracker~example')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=False, nargs='*'),

        { 'names': ('--sort', '-s'),
          'default_description': "current value of 'sort.torrents' setting",
          'description': ('Comma-separated list of sort orders '
                          "(see SORT ORDERS section)") },

        { 'names': ('--columns', '-c'),
          'default_description': "current value of 'columns.torrents' setting",
          'description': ('Comma-separated list of column names '
                          "(see COLUMNS section)") },
    )

    from ...views.torrent import COLUMNS
    from ...client.sorters.torrent import TorrentSorter
    more_sections = {
        'COLUMNS': make_COLUMNS_doc(COLUMNS, '--columns', 'columns.torrents'),
        'SORT ORDERS': make_SORT_ORDERS_doc(TorrentSorter, '--sort', 'sort.torrents'),
        'SCRIPTING': make_SCRIPTING_doc(name),
    }

    cfg = ExpectedResource

    async def run(self, TORRENT_FILTER, sort, columns):
        sort = self.cfg['sort.torrents'] if sort is None else sort
        columns = self.cfg['columns.torrents'] if columns is None else columns
        try:
            columns = self.get_torrent_columns(columns)
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=True,
                                           discover_torrent=False)
            sort = self.get_torrent_sorter(sort)
        except ValueError as e:
            raise CmdError(e)
        else:
            log.debug('Listing %s torrents sorted by %s', tfilter, sort)
            if asyncio.iscoroutinefunction(self.make_torrent_list):
                await self.make_torrent_list(tfilter, sort, columns)
            else:
                self.make_torrent_list(tfilter, sort, columns)


class TorrentSummaryCmdbase(mixin.get_single_torrent, metaclass=InitCommand):
    name = 'summary'
    aliases = ('info', 'details')
    provides = set()
    category = 'torrent'
    description = 'Display detailed torrent information'
    usage = ('summary',
             'summary <TORRENT FILTER>')
    examples = ('summary id=71',)
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),
    )
    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True,
                                           prefer_focused=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            torrent = await self.get_single_torrent(tfilter, keys=('id', 'name'))
            if not torrent:
                raise CmdError()
            else:
                log.debug('Showing summary of torrent %r: %r', tfilter, torrent)
                if asyncio.iscoroutinefunction(self.display_summary):
                    await self.display_summary(torrent['id'])
                else:
                    self.display_summary(torrent['id'])


class AddTorrentsCmdbase(metaclass=InitCommand):
    name = 'add'
    aliases = ('download','get')
    provides = set()
    category = 'torrent'
    description = 'Download torrents'
    usage = ('add [<OPTIONS>] <TORRENT> <TORRENT> <TORRENT> ...',)
    examples = ('add 72d7a3179da3de7a76b98f3782c31843e3f818ee',
                'add --stopped http://example.org/something.torrent')
    argspecs = (
        { 'names': ('TORRENT',), 'nargs': '+',
          'description': 'Link or path to torrent file, magnet link or info hash' },

        { 'names': ('--stopped','-s'), 'action': 'store_true',
          'description': 'Do not start downloading the added torrent(s)' },

        { 'names': ('--path','-p'),
          'description': 'Custom download directory for added torrent(s)' },
    )
    srvapi = ExpectedResource

    async def run(self, TORRENT, stopped, path):
        success = True
        force_torrentlist_update = False
        for source in TORRENT:
            response = await self.make_request(self.srvapi.torrent.add(source, stopped=stopped, path=path))
            success = success and response.success
            force_torrentlist_update = force_torrentlist_update or success

        # Update torrentlist AFTER all 'add' requests
        if force_torrentlist_update and hasattr(self, 'polling_frenzy'):
            self.polling_frenzy()

        if not success:
            raise CmdError()


class MoveTorrentsCmdbase(metaclass=InitCommand):
    name = 'move'
    aliases = ('mv',)
    provides = set()
    category = 'torrent'
    description = "Change torrents' location"
    usage = ('move <PATH>',
             'move <TORRENT FILTER> <PATH>')
    examples = ('move ./new/path',
                'move size>50G /path/to/lots/of/storage')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),

        {'names': ('PATH',),
         'description': ('Move the specified torrent(s) to this directory.  If PATH is relative '
                         '(i.e. does not start with "/"), it is relative to the value of the '
                         'setting "srv.path.complete".  That means "." is the download path.')},
    )
    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, PATH):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            response = await self.make_request(self.srvapi.torrent.move(tfilter, PATH),
                                               polling_frenzy=True)
            if not response.success:
                raise CmdError()


class RenameTorrentCmdbase(metaclass=InitCommand):
    name = 'rename'
    aliases = ('rn',)
    provides = set()
    category = 'torrent'
    description = 'Rename a torrent or one of its files or directories'
    usage = ('rename <NEW>',
             'rename <CURRENT> <NEW>')
    examples = ('rename "A Better Name"',
                'rename id=123 Foo',
                'rename id=123/some/file new_file_name')
    argspecs = (
        { 'names': ('CURRENT',), 'nargs': '?',
          'description': ('Torrent filter expression that matches exactly one torrent, '
                          'optionally followed by a "/" and the relative path to a '
                          'file or directory in the torrent'),
          'default_description': 'Focused torrent, file or directory in the TUI' },

        {'names': ('NEW',),
         'description': ('New name of the torrent, file or directory specified by CURRENT '
                         '(must not contain "/" or be "." or "..")')},
    )
    srvapi = ExpectedResource

    async def run(self, CURRENT, NEW):
        # Autodetect current path
        if not CURRENT:
            path = self.get_focused_path_in_torrent()
            if path:
                # path is <torrent name>/<relative/path/to/file/in/torrent>
                CURRENT = path

        # Split filter from current path
        if CURRENT and '/' in CURRENT:
            FILTER, PATH = CURRENT.split('/', maxsplit=1)
        else:
            FILTER, PATH = CURRENT, None

        try:
            tfilter = self.select_torrents(FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            response = await self.make_request(
                self.srvapi.torrent.torrents(tfilter, keys=('id',)),
                quiet=True)
            if not response.success:
                raise CmdError()
            elif len(response.torrents) > 1:
                raise CmdError('%s matches more than one torrent' % tfilter)
            else:
                tid = response.torrents[0]['id']
                response = await self.make_request(
                    self.srvapi.torrent.rename(tid, path=PATH, new_name=NEW),
                    polling_frenzy=True)
                if not response.success:
                    raise CmdError()


class RemoveTorrentsCmdbase(metaclass=InitCommand):
    name = 'remove'
    aliases = ('rm', 'delete')
    provides = set()
    category = 'torrent'
    description = 'Remove torrents'
    usage = ('remove [<OPTIONS>]',
             'remove [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('remove',
                'remove "stupid torrent" silly\ torrent and_this_torrent',
                'remove -d "unwanted torrent"')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),

        { 'names': ('--delete-files','-d'), 'action': 'store_true',
          'description': 'Delete any downloaded files' },

        { 'names': ('--force','-f'), 'action': 'store_true',
          'description': ('Ignore remove.max-hits setting: Remove all '
                          'matching torrents instead of asking for confirmation '
                          'if the number of matches exceeds remove.max-hits')},
    )
    srvapi = ExpectedResource
    cfg = ExpectedResource

    async def run(self, TORRENT_FILTER, delete_files, force):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            async def do_remove(tfilter=tfilter, delete_files=delete_files):
                response = await self.make_request(
                    self.srvapi.torrent.remove(tfilter, delete=delete_files),
                    polling_frenzy=True)
                if not response.success:
                    raise CmdError()

            async def do_keep(tfilter=tfilter):
                self.error(('Keeping %s torrents: Too many hits ' % tfilter) +
                           '(use --force or increase remove.max-hits setting)')

            response = await self.srvapi.torrent.torrents(tfilter, keys=('id',))
            hits = len(response.torrents)
            success = hits > 0
            if force or self.cfg['remove.max-hits'] < 0 or hits < self.cfg['remove.max-hits']:
                return await do_remove()
            else:
                await self.show_list_of_hits(tfilter)
                if hits > 0:
                    question = 'Are you sure you want to remove %d torrent%s' % (
                        hits, '' if hits == 1 else 's')
                    if delete_files:
                        question += ' and their files'
                    question += '?'
                    success = await self.ask_yes_no(question, yes=do_remove, no=do_keep,
                                                    after=self.remove_list_of_hits)
                if not success:
                    raise CmdError()


# Argument definitions that are shared between commands
ARGSPEC_TOGGLE = {
    'names': ('--toggle','-t'), 'action': 'store_true',
    'description': ('Start TORRENT if stopped and vice versa')
}

class StartTorrentsCmdbase(metaclass=InitCommand):
    name = 'start'
    aliases = ()
    provides = set()
    category = 'torrent'
    description = 'Start downloading torrents'
    usage = ('start [<OPTIONS>]',
             'start [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('start',
                "start 'night of the living dead' Metropolis",
                'start ubuntu --force')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),

        { 'names': ('--force','-f'), 'action': 'store_true',
          'description': 'Ignore download queue' },

        ARGSPEC_TOGGLE,
    )
    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, toggle, force):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            if toggle:
                response = await self.make_request(
                    self.srvapi.torrent.toggle_stopped(tfilter, force=force),
                    polling_frenzy=True)
            else:
                response = await self.make_request(
                    self.srvapi.torrent.start(tfilter, force=force),
                    polling_frenzy=True)
            if not response.success:
                raise CmdError()


class StopTorrentsCmdbase(metaclass=InitCommand):
    name = 'stop'
    aliases = ('pause',)
    provides = set()
    category = 'torrent'
    description = 'Stop downloading torrents'
    usage = ('stop [<OPTIONS>]',
             'stop [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('stop',
                'stop "night of the living dead" idle',
                'stop --toggle ubuntu')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),
        ARGSPEC_TOGGLE,
    )
    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, toggle):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            if toggle:
                response = await self.make_request(
                    self.srvapi.torrent.toggle_stopped(tfilter),
                    polling_frenzy=True)
            else:
                response = await self.make_request(
                    self.srvapi.torrent.stop(tfilter),
                    polling_frenzy=True)
            if not response.success:
                raise CmdError()


class VerifyTorrentsCmdbase(metaclass=InitCommand):
    name = 'verify'
    aliases = ('check',)
    provides = set()
    category = 'torrent'
    description = 'Verify downloaded torrent data'
    usage = ('verify [<OPTIONS>]',
             'verify [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('verify',
                'verify debian')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),
    )
    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            response = await self.make_request(self.srvapi.torrent.verify(tfilter),
                                               polling_frenzy=False)
            if not response.success:
                raise CmdError()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filter sequences of Torrents, TorrentFiles by arbitrary criteria"""

from ...logging import make_logger
log = make_logger(__name__)

import operator
import re
from collections import abc
from itertools import zip_longest
from ..ttypes import (Timedelta, Timestamp)
from ..utils import const

def timestamp_or_timedelta(string, default_sign=1):
    try:
        return Timestamp.from_string(string)
    except ValueError:
        delta = Timedelta.from_string(string)
        # Apply default_sign if no sign was given explicitly
        string = string.strip()
        if (not string.startswith('in') and
            not string.endswith('ago') and
            string[0] not in ('+', '-')):
            delta = Timedelta(delta * default_sign)
        return delta

def time_filter(torrent_value, op, user_value):
    if not torrent_value.is_known:
        return False

    type_torrent_value = type(torrent_value)
    type_user_value = type(user_value)

    if type_torrent_value is Timestamp:
        if type_user_value is Timestamp:
            return op(torrent_value, user_value)
        elif type_user_value is Timedelta:
            return _compare_timedelta(torrent_value.timedelta, op,  user_value)

    elif type_torrent_value is Timedelta:
        if type_user_value is Timedelta:
            return _compare_timedelta(torrent_value, op, user_value)
        elif type_user_value is Timestamp:
            return _compare_timedelta(torrent_value, op, user_value.timedelta)

    raise RuntimeError('cannot compare {torrent_value!r} with {user_value!}')

def _compare_timedelta(delta_torrent, op, delta_user):
    if delta_user <= 0:
        # User's time delta is in the past - ignore future times
        if delta_torrent > 0:
            return False
    elif delta_user > 0:
        # User's time delta is in the future - ignore past times
        if delta_torrent <= 0:
            return False
    return op(abs(delta_torrent), abs(delta_user))


def limit_rate_filter(limit, op, user_value):
    # `limit` may be a number or const.UNLIMITED.
    # `user_value` may be a number or a Bool (which is not a derivative of the
    # built-in `bool`).
    # If `user_value` is a Bool, True means 'limited' and False means
    # 'unlimited'.

    if isinstance(user_value, (int, float)):
        # This works because const.UNLIMITED behaves like `float('inf')`.
        return op(limit, user_value)
    else:
        if not user_value:
            return op(limit, const.UNLIMITED)

        # `user_value` is 'limited'.
        elif op.__name__ in ('gt', 'ge'):
            # 'greater than limited' -> 'equal to unlimited'
            return limit == const.UNLIMITED
        elif op.__name__ in ('eq', 'ne'):
            # 'equal/unequal to limited' -> 'not equal/unequal to unlimited'
            return not op(limit, const.UNLIMITED)
        elif op.__name__ == 'lt':
            # 'lower than limited' -> No idea what to do here.
            return False
        elif op.__name__ == 'le':
            # Ignore the 'lower' in 'lower or equal';
            # 'equal to limited' -> ' unequal to unlimited'
            return limit != const.UNLIMITED
        else:
            # That should be all possible operators, but to avoid any
            # tracebacks, whatever is left gets regular treatment
            return op(limit, user_value)


class BoolFilterSpec():
    """Boolean filter specification"""

    def __init__(self, func, needed_keys=(), aliases=(), description='No description'):
        self.filter_function = func
        self.needed_keys = needed_keys
        self.aliases = aliases
        self.description = description

class CmpFilterSpec(BoolFilterSpec):
    """Comparative filter specification"""

    def __init__(self, *args, value_type, value_convert=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.value_type = value_type
        self.value_convert = value_convert or value_type

    def make_filter_func(self, operator, value):
        def func(obj):
            return self.filter_function(obj, operator, value)
        return func

def make_cmp_filter(types, key, description, aliases=(), value_convert=None):
    def filterfunc(obj, op, val, key=key):
        return op(obj[key], val)

    kwargs = {'description'   : description,
              'needed_keys'   : (key,),
              'aliases'       : aliases,
              'value_type'    : types[key],
              'value_convert' : value_convert}

    if kwargs['value_convert'] is None and hasattr(kwargs['value_type'], 'from_string'):
        kwargs['value_convert'] = kwargs['value_type'].from_string

    return CmpFilterSpec(filterfunc, **kwargs)


def _unquote(string):
    if len(string) <= 2:
        return string

    firstchar, lastchar = string[0], string[-1]
    if (firstchar == "'" and lastchar == "'" or
          firstchar == '"' and lastchar == '"'):
        return string[1:-1]

    return string


class Filter():
    """Match sequences of objects against a single filter"""

    _OPERATORS = {
        '=': operator.__eq__, '~': operator.__contains__,
        '>': operator.__gt__, '<': operator.__lt__,
        '>=': operator.__ge__, '<=': operator.__le__,
    }
    _OP_CHARS = ''.join(_OPERATORS)
    _OP_LIST = '(?:' + '|'.join(sorted(_OPERATORS, key=lambda op: len(op), reverse=True)) + ')'
    _INVERT_CHAR = '!'
    _FILTER_REGEX = re.compile(r'^'
                               r'(?P<invert1>' + _INVERT_CHAR + '?)'
                               r'(?P<name>[^' + _OP_CHARS+_INVERT_CHAR + ']*)'
                               r'(?P<invert2>' + _INVERT_CHAR + '?)'
                               r'(?P<op>' + _OP_LIST + '|)'
                               r'(?P<value>.*)$')

    DEFAULT_FILTER = None

    BOOLEAN_FILTERS = {}
    COMPARATIVE_FILTERS = {}

    @classmethod
    def _check_value(cls, name, value, op):
        """
        Convert `value` to correct type for comparative filter `name`

        Also ensure operator `op` is compatible with `value`.

        Raises ValueError
        """
        if name not in cls.COMPARATIVE_FILTERS:
            return value

        # Convert value to proper type
        target_type = cls.COMPARATIVE_FILTERS[name].value_type
        if type(value) is not target_type:
            converter = cls.COMPARATIVE_FILTERS[name].value_convert
            try:
                value = converter(value)
            except ValueError:
                raise ValueError('Invalid value for filter {!r}: {!r}'.format(name, value))

        # Test if target_type supports operator
        try:
            cls._OPERATORS[op](value, value)
        except TypeError as e:
            raise ValueError('Invalid operator for filter {!r}: {}'.format(name, op))

        return value

    @classmethod
    def _resolve_alias(cls, name):
        if not hasattr(cls, '_aliases'):
            aliases = {}
            for fseq in (cls.BOOLEAN_FILTERS, cls.COMPARATIVE_FILTERS):
                for fname,f in fseq.items():
                    for a in f.aliases:
                        aliases[a] = fname
            cls._aliases = aliases
        return cls._aliases.get(name, name)

    def __init__(self, filter_str=''):
        # name: Name of filter (user-readable string)
        # invert: Whether to invert filter (bool)
        # op: Comparison operator as string (see _OPERATORS)
        # value: User-given value as string (will be converted to proper type)
        name, invert, op, value = (None, False, None, None)
        filter_str = filter_str.strip()
        if filter_str != '':
            match = self._FILTER_REGEX.fullmatch(filter_str)
            if match is None:
                raise ValueError('Invalid filter: {!r}'.format(filter_str))
            else:
                name = match.group('name')
                op = match.group('op') or None
                invert = bool(match.group('invert1')) ^ bool(match.group('invert2'))
                value = match.group('value')
                value = _unquote(value.strip(' ')) if value else None

        # No operator but a value doesn't make any sense
        if op is None and value is not None:
            raise ValueError('Malformed filter expression: {!r}'.format(filter_str))

        if name is not None:
            name = name.strip()

        if name is None:
            # No filter_str provided
            name = 'all'
        elif name is '':
            # No filter name is given, but a value and maybe an operator.
            name = self.DEFAULT_FILTER
        else:
            name = self._resolve_alias(name)

        # Make sure value has the correct type and operator is compatible
        if value is not None:
            value = self._check_value(name, value, op)

        log.debug('Parsed filter %r: name=%r, invert=%r, op=%r, value=%r',
                  filter_str, name, invert, op, value)

        # Filter that doesn't use value argument
        if name in self.BOOLEAN_FILTERS:
            if op or value:
                raise ValueError('Boolean filter does not accept any values: {} '.format(filter_str))
            f = self.BOOLEAN_FILTERS[name]
            self._filter_func = f.filter_function
            self._needed_keys = f.needed_keys

        # Filter that needs an argument
        elif name in self.COMPARATIVE_FILTERS:
            f = self.COMPARATIVE_FILTERS[name]
            self._needed_keys = f.needed_keys
            if op is None and value is None:
                # Abuse comparative filter as boolean filter
                # (e.g. 'peers-connected' matches torrents with peers-connected!=0)
                self._filter_func = lambda obj, keys=f.needed_keys: all(bool(obj[key]) for key in keys)
            elif op is None:
                ops = '[' + '|'.join(sorted(self._OPERATORS)) + ']'
                raise ValueError('Missing operator and value: {} {} ...'.format(name, ops))
            elif value is None:
                raise ValueError('Missing value: {} ...'.format(filter_str))
            else:
                self._filter_func = f.make_filter_func(self._OPERATORS[op], value)

        elif value is op is None and self.DEFAULT_FILTER is not None:
            # `name` is no known filter - default to DEFAULT_FILTER with operator '~'.
            value = _unquote(name)
            op = '~'
            name = self.DEFAULT_FILTER
            if name in self.BOOLEAN_FILTERS:
                f = self.BOOLEAN_FILTERS[name]
                self._filter_func = f.filter_function
            elif name in self.COMPARATIVE_FILTERS:
                f = self.COMPARATIVE_FILTERS[name]
                self._filter_func = f.make_filter_func(self._OPERATORS[op], value)
            else:
                raise RuntimeError('Default filter {!r} does not exist: {!r}'
                                   .format(name, ', '.join(tuple(self.BOOLEAN_FILTERS) +
                                                           tuple(self.COMPARATIVE_FILTERS))))
            self._needed_keys = f.needed_keys

        else:
            raise ValueError('Invalid filter name: {!r}'.format(name))

        self._name, self._invert, self._op, self._value = name, invert, op, value
        self._hash = hash((name, invert, op, value))

    def apply(self, objs, invert=False, key=None):
        """Yield matching objects or `key` of each matching object"""
        invert = self._invert ^ bool(invert)  # xor
        wanted = self._filter_func
        for obj in objs:
            if wanted(obj) ^ invert:
                yield obj if key is None else obj[key]

    def match(self, obj):
        """Return True if `obj` matches, False otherwise"""
        return self._filter_func(obj) ^ self._invert

    def __str__(self):
        if self._name is None:
            return 'all'
        elif self._value is None:
            return ('!' if self._invert else '') + self._name
        else:
            name = self._name if self._name != self.DEFAULT_FILTER else ''
            op = ('!' if self._invert else '') + self._op
            val = str(self._value)
            if val == '' or val[0] == ' ' or val[-1] == ' ':
                val = repr(val)
            return name + op + val

    @property
    def needed_keys(self):
        return self._needed_keys

    def __eq__(self, other):
        if isinstance(other, type(self)):
            for attr in ('_name', '_value', '_invert', '_op'):
                if getattr(self, attr) != getattr(other, attr):
                    return False
            return True
        else:
            return NotImplemented

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, str(self))

    def __hash__(self):
        return self._hash



class FilterChain():
    """One or more filters combined with AND and OR operators"""

    filterclass = None
    _op_regex = re.compile(r'(?<!\\)([&|])')

    def __init__(self, filters=''):
        if not isinstance(self.filterclass, type) or not issubclass(self.filterclass, Filter):
            raise RuntimeError('Attribute "filterclass" must be set to a Filter class, not {!r}'
                               .format(self.filterclass))

        if isinstance(filters, str):  # Because str is also instance of abc.Sequence
            pass
        elif isinstance(filters, abc.Sequence) and all(isinstance(f, str) for f in filters):
            filters = '|'.join(filters)
        elif not isinstance(filters, str):
            raise TypeError('filters must be string or sequence of strings, not {}: {!r}'
                            .format(type(filters).__name__, filters))

        # self._filterchains is a tuple of tuples.  Each inner tuple combines
        # filters with AND.  The outer tuple combines the inner, AND-combined
        # tuples with OR.
        parts = tuple(part for part in self._op_regex.split(filters) if part is not '')
        if len(parts) < 1:
            self._filterchains = ()
        else:
            if parts[0] in '&|':
                raise ValueError('Filter can\'t start with operator: {!r}'.format(parts[0]))
            elif parts[-1] in '&|':
                raise ValueError('Filter can\'t end with operator: {!r}'.format(parts[-1]))

            filters = []
            ops = []
            expect = 'filter'
            nofilter = self.filterclass()
            for i,part in enumerate(parts):
                if expect is 'filter':
                    if part not in '&|':
                        # Remove backslashes from escaped operators
                        part = part.replace('\&', '&').replace('\|', '|')
                        f = self.filterclass(part)
                        if f == nofilter:
                            # part is something like 'all' or '*' - this
                            # disables all other filters
                            filters = []
                            ops = []
                            break
                        else:
                            filters.append(f)
                            expect = 'operator'
                            continue
                elif expect is 'operator':
                    if part in '&|':
                        ops.append(part)
                        expect = 'filter'
                        continue
                raise ValueError('Consecutive operators: {!r}'.format(''.join(parts[i-2:i+2])))

            if filters:
                fchain = [[]]
                for filter,op in zip_longest(filters, ops):
                    fchain[-1].append(filter)
                    if op is '|':
                        fchain.append([])
                self._filterchains = tuple(tuple(x) for x in fchain)
            else:
                self._filterchains = ()

    def apply(self, objects):
        """Yield matching objects from iterable `objects`"""
        if self._filterchains:
            yield from filter(self.match, objects)
        else:
            yield from objects

    def match(self, obj):
        """Whether `obj` matches this filter chain"""
        # All filters in an AND_chain must match for the AND_chain to
        # match.  At least one AND_chain must match.
        if len(self._filterchains) < 1:
            return True
        else:
            return any(all(f.match(obj) for f in AND_chain)
                       for AND_chain in self._filterchains)

    @property
    def needed_keys(self):
        """The object keys needed for filtering"""
        keys = set()
        for chain in self._filterchains:
            for filter in chain:
                keys.update(filter.needed_keys)
        return tuple(keys)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            # Compare sets because order doesn't matter (foo&bar|baz is the
            # same as baz|bar&foo). Use frozensets because sets are not
            # hashable.
            self_fc_sets = set(frozenset(x) for x in self._filterchains)
            other_fc_sets = set(frozenset(x) for x in other._filterchains)
            return self_fc_sets == other_fc_sets

    def __str__(self):
        if len(self._filterchains) < 1:
            return 'all'
        else:
            OR_chains = []
            for AND_chain in self._filterchains:
                OR_chains.append('&'.join(str(f) for f in AND_chain))
            return '|'.join(OR_chains)

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, str(self))

    def __and__(self, other):
        cls = type(self)
        if not isinstance(other, cls):
            return NotImplemented
        else:
            return cls(str(self) + '&' + str(other))

    def __or__(self, other):
        cls = type(self)
        if not isinstance(other, cls):
            return NotImplemented
        else:
            return cls(str(self) + '|' + str(other))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from ..base import file as base
from . import _mixin as mixin
from .. import (CmdError, ExpectedResource)
from ._table import (print_table, TERMSIZE)

import os


class ListFilesCmd(base.ListFilesCmdbase,
                   mixin.make_request, mixin.select_torrents, mixin.select_files,
                   mixin.only_supported_columns):
    provides = {'cli'}
    srvapi = ExpectedResource

    async def make_file_list(self, tfilter, ffilter, columns):
        response = await self.make_request(
            self.srvapi.torrent.torrents(tfilter, keys=('name', 'files')),
            quiet=True)
        torrents = response.torrents

        if len(torrents) < 1:
            raise CmdError()

        filelist = []
        for torrent in sorted(torrents, key=lambda t: t['name'].lower()):
            files, filtered_count = self._flatten_tree(torrent['files'], ffilter)
            filelist.extend(files)

        if filelist:
            from ...views.file import COLUMNS as FILE_COLUMNS
            # Remove columns that aren't supported by CLI interface (e.g. 'marked')
            columns = self.only_supported_columns(columns, FILE_COLUMNS)
            print_table(filelist, columns, FILE_COLUMNS)
        else:
            if str(tfilter) != 'all':
                raise CmdError('No matching files in %s torrents: %s' % (tfilter, ffilter))
            else:
                raise CmdError('No matching files: %s' % (ffilter))

    def _flatten_tree(self, files, ffilter=None, _indent_level=0):
        """
        Return list of rows for `print_table`

        `files` must be a nested mapping tree (i.e. TorrentFileTree).
        `ffilter` must be a TorrentFileFilter instance or None.
        """
        if TERMSIZE.columns is None:
            def indent(node):
                node['name'] = node['path-absolute']
        else:
            def indent(node):
                node['name'] = '%s%s' % ('  '*(_indent_level), node['name'])

        from ...views.file import TorrentFileDirectory
        flist = []
        filtered_count = 0
        for key,value in sorted(files.items(), key=lambda pair: pair[0].lower()):
            if value.nodetype == 'leaf':
                if ffilter is None or ffilter.match(value):
                    filenode = dict(value)  # Copy original TorrentFile
                    indent(filenode)
                    flist.append(filenode)
                else:
                    filtered_count += 1

            elif value.nodetype == 'parent':
                sub_flist, sub_filtered_count = self._flatten_tree(value, ffilter, _indent_level+1)
                if TERMSIZE.columns is not None:
                    dirnode = TorrentFileDirectory(key, value, sub_filtered_count)
                    indent(dirnode)
                    flist.append(dirnode)
                flist.extend(sub_flist)

        return flist, filtered_count


class PriorityCmd(base.PriorityCmdbase,
                  mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'cli'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from . import (SortSpecBase, SorterBase)

class _SortSpec(SortSpecBase):
    def __init__(self, *args, description='', **kwargs):
        description = 'Sort trackers by %s' % description
        super().__init__(*args, description=description, **kwargs)


class TorrentTrackerSorter(SorterBase):
    DEFAULT_SORT = 'domain'
    SORTSPECS = {
        'torrent':         _SortSpec(lambda t: t['tname'].lower(),
                                     description='torrent name'),
        'tier':            _SortSpec(lambda t: t['tier'],
                                     description='tier number'),
        'domain':          _SortSpec(lambda t: t['domain'],
                                     aliases=('dom',),
                                     description='domain from announce URL'),
        'status':          _SortSpec(lambda t: t['status'],
                                     aliases=('st',),
                                     description='tracker status'),
        'error':           _SortSpec(lambda t: t['error'],
                                     aliases=('err',),
                                     description='error message'),
        'downloads':       _SortSpec(lambda t: t['count-downloads'],
                                     aliases=('dns',),
                                     description='number of known downloads'),
        'leeches':         _SortSpec(lambda t: t['count-leeches'],
                                     aliases=('lcs',),
                                     description='number of known downloading peers'),
        'seeds':           _SortSpec(lambda t: t['count-seeds'],
                                     aliases=('sds',),
                                     description='number of known seeding peers'),
        'last-announce':   _SortSpec(lambda t: t['time-last-announce'],
                                     aliases=('lan',),
                                     description='last time the torrent was successfully announce'),
        'next-announce':   _SortSpec(lambda t: t['time-next-announce'],
                                     aliases=('nan',),
                                     description='next time the torrent is announced'),
        'last-scrape':     _SortSpec(lambda t: t['time-last-scrape'],
                                     aliases=('lsc',),
                                     description='last time the torrent was successfully scrape'),
        'next-scrape':     _SortSpec(lambda t: t['time-next-scrape'],
                                     aliases=('nsc',),
                                     description='next time the torrent is scraped'),
    }



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering Torrents by their values"""

from ...logging import make_logger
log = make_logger(__name__)

from . import (BoolFilterSpec, CmpFilterSpec, make_cmp_filter, Filter, FilterChain)
from . import (timestamp_or_timedelta, time_filter, limit_rate_filter)
from ..utils import (Bandwidth, BoolOrBandwidth, convert)


from ..ttypes import TYPES as VALUETYPES
def _make_cmp_filter(*args, **kwargs):
    return make_cmp_filter(VALUETYPES, *args, **kwargs)

def _desc(text):
    if text.startswith('...'):
        text = 'Match VALUE against ' + text[4:]
    return text


from ..ttypes import Status
_STATUS_VERIFY    = Status.VERIFY
_STATUS_DOWNLOAD  = Status.DOWNLOAD
_STATUS_UPLOAD    = Status.UPLOAD
_STATUS_INIT      = Status.INIT
_STATUS_CONNECTED = Status.CONNECTED
_STATUS_ISOLATED  = Status.ISOLATED
_STATUS_QUEUED    = Status.QUEUED
_STATUS_SEED      = Status.SEED
_STATUS_IDLE      = Status.IDLE
_STATUS_STOPPED   = Status.STOPPED

class SingleTorrentFilter(Filter):
    DEFAULT_FILTER = 'name'

    # Filters without arguments
    BOOLEAN_FILTERS = {
        'all': BoolFilterSpec(
            lambda t: True,
            aliases=('*',),
            description='All torrents',
            needed_keys=()),
        'complete': BoolFilterSpec(
            lambda t: t['%downloaded'] >= 100,
            aliases=('cmp',),
            description='Torrents with all wanted files downloaded',
            needed_keys=('%downloaded',)),
        'incomplete': BoolFilterSpec(
            lambda t: t['%downloaded'] < 100,
            aliases=('inc',),
            description='Torrents with some wanted files not fully downloaded',
            needed_keys=('%downloaded',)),
        'stopped': BoolFilterSpec(
            lambda t: _STATUS_STOPPED in t['status'],
            aliases=('stp',),
            description='Torrents not allowed to up- or download',
            needed_keys=('status',)),
        'active': BoolFilterSpec(
            lambda t: t['peers-connected'] > 0 or _STATUS_VERIFY in t['status'],
            description='Torrents connected to peers or being verified',
            needed_keys=('peers-connected', 'status')),
        'uploading': BoolFilterSpec(
            lambda t: t['rate-up'] > 0,
            aliases=('upg',),
            description='Torrents using upload bandwidth',
            needed_keys=('rate-up',)),
        'downloading': BoolFilterSpec(
            lambda t: t['rate-down'] > 0,
            aliases=('dng',),
            description='Torrents using download bandwidth',
            needed_keys=('rate-down',)),
        'verifying': BoolFilterSpec(
            lambda t: _STATUS_VERIFY in t['status'],
            aliases=('vfg',),
            description='Torrents being verified or queued for verification',
            needed_keys=('status',)),
        'leeching': BoolFilterSpec(
            lambda t: t['%downloaded'] < 100 and _STATUS_STOPPED not in t['status'],
            aliases=('lcg',),
            description='Unstopped torrents downloading or waiting for seeds',
            needed_keys=('%downloaded', 'status')),
        'seeding': BoolFilterSpec(
            lambda t: t['%downloaded'] >= 100 and _STATUS_STOPPED not in t['status'],
            aliases=('sdg',),
            description='Unstopped torrents with all wanted files downloaded',
            needed_keys=('%downloaded', 'status')),
        'idle': BoolFilterSpec(
            lambda t: (_STATUS_IDLE in t['status'] and
                       _STATUS_STOPPED not in t['status']),
            description='Unstopped torrents not using any bandwidth',
            needed_keys=('status',)),
        'isolated': BoolFilterSpec(
            lambda t: _STATUS_ISOLATED in t['status'],
            aliases=('isl',),
            description='Torrents that cannot discover new peers in any way',
            needed_keys=('status',)),
        'private': BoolFilterSpec(
            lambda t: t['private'],
            aliases=('prv',),
            description='Torrents connectable through trackers only',
            needed_keys=('private',)),
        'public': BoolFilterSpec(
            lambda t: not t['private'],
            aliases=('pbl',),
            description='Torrents connectable through DHT and/or PEX',
            needed_keys=('private',)),
    }

    # Filters with arguments
    COMPARATIVE_FILTERS = {
        'id'              : _make_cmp_filter('id',
                                             _desc('... ID')),
        'name'            : _make_cmp_filter('name',
                                             _desc('... name'),
                                             aliases=('n',)),
        'comment'         : _make_cmp_filter('comment',
                                             _desc('... comment'),
                                             aliases=('cmnt',)),
        'path'            : _make_cmp_filter('path',
                                             _desc('... full path to download directory'),
                                             aliases=('dir',)),
        'error'           : _make_cmp_filter('error',
                                             _desc('... error message'),
                                             aliases=('err',)),
        'uploaded'        : _make_cmp_filter('size-uploaded',
                                             _desc('... number of uploaded bytes'),
                                             aliases=('up',)),
        'downloaded'      : _make_cmp_filter('size-downloaded',
                                             _desc('... number of downloaded bytes'),
                                             aliases=('dn',)),
        '%downloaded'     : _make_cmp_filter('%downloaded',
                                             _desc('... percentage of downloaded bytes'),
                                             aliases=('%dn',)),
        'size'            : _make_cmp_filter('size-final',
                                             _desc('... combined size of all wanted files'),
                                             aliases=('sz',),
                                             value_convert=convert.size),
        'peers'           : _make_cmp_filter('peers-connected',
                                             _desc('... number of connected peers'),
                                             aliases=('prs',)),
        'seeds'           : _make_cmp_filter('peers-seeding',
                                             _desc('... largest number of seeds reported by any tracker'),
                                             aliases=('sds',)),
        'ratio'           : _make_cmp_filter('ratio',
                                             _desc('... uploaded/downloaded ratio'),
                                             aliases=('rto',)),
        'rate-up'         : _make_cmp_filter('rate-up',
                                             _desc('... upload rate'),
                                             aliases=('rup',),
                                             value_convert=Bandwidth),
        'rate-down'       : _make_cmp_filter('rate-down',
                                             _desc('... download rate'),
                                             aliases=('rdn',),
                                             value_convert=Bandwidth),

        'limit-rate-up': CmpFilterSpec(
            lambda t, op, v: limit_rate_filter(t['limit-rate-up'], op, v),
            aliases=('lrup',),
            description=_desc('... upload rate limit'),
            needed_keys=('limit-rate-up',),
            value_type=BoolOrBandwidth,
        ),

        'limit-rate-down': CmpFilterSpec(
            lambda t, op, v: limit_rate_filter(t['limit-rate-down'], op, v),
            aliases=('lrdn',),
            description=_desc('... download rate limit'),
            needed_keys=('limit-rate-down',),
            value_type=BoolOrBandwidth,
        ),

        'tracker': CmpFilterSpec(
            lambda t, op, v: any(op(tracker['url-announce'].domain, v)
                                 for tracker in t['trackers']),
            aliases=('trk',),
            description=_desc('... domain of the announce URL of trackers'),
            needed_keys=('trackers',),
            value_type=str,
        ),

        'eta': CmpFilterSpec(
            lambda t, op, v: time_filter(t['timespan-eta'], op, v),
            description=_desc('... estimated time for torrent to finish'),
            needed_keys=('timespan-eta',),
            value_type=VALUETYPES['timespan-eta'],
            value_convert=timestamp_or_timedelta,
        ),

        'created': CmpFilterSpec(
            lambda t, op, v: time_filter(t['time-created'], op, v),
            aliases=('tcrt',),
            description=_desc('... time torrent was created'),
            needed_keys=('time-created',),
            value_type=VALUETYPES['time-created'],
            value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
        ),
        'added': CmpFilterSpec(
            lambda t, op, v: time_filter(t['time-added'], op, v),
            aliases=('tadd',),
            description=_desc('... time torrent was added'),
            needed_keys=('time-added',),
            value_type=VALUETYPES['time-added'],
            value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
        ),
        'started': CmpFilterSpec(
            lambda t, op, v: time_filter(t['time-started'], op, v),
            aliases=('tsta',),
            description=_desc('... last time torrent was started'),
            needed_keys=('time-started',),
            value_type=VALUETYPES['time-started'],
            value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
        ),
        'activity': CmpFilterSpec(
            lambda t, op, v: time_filter(t['time-activity'], op, v),
            aliases=('tact',),
            description=_desc('... last time torrent was active'),
            needed_keys=('time-activity',),
            value_type=VALUETYPES['time-activity'],
            value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
        ),
        'completed': CmpFilterSpec(
            lambda t, op, v: time_filter(t['time-completed'], op, v),
            aliases=('tcmp',),
            description=_desc('... time all wanted files where downloaded'),
            needed_keys=('time-completed',),
            value_type=VALUETYPES['time-completed'],
            value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
        ),
    }


class TorrentFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = SingleTorrentFilter



# Create long_description from README.org if possible
long_description = ''
try:
    import pypandoc
    import os
    if os.path.exists('README.org'):
        long_description = pypandoc.convert('README.org', 'rst')
        import restructuredtext_lint as rst
        rst.lint(long_description)
except ImportError:
    pass

# Load __version__ variable without importing the whole stig module
with open('stig/__version__.py') as f:
    exec(f.read())

from setuptools import setup, find_packages
setup(
    name             = 'stig',
    version          = __version__,
    license          = 'GPLv3+',
    author           = 'Random User',
    author_email     = 'rndusr@posteo.de',

    description      = 'TUI and CLI client for the Transmission daemon',
    long_description = long_description,

    url              = 'https://github.com/rndusr/stig',
    keywords         = 'bittorrent torrent transmission',

    packages         = find_packages(),
    package_data     = {'stig': ['settings/default.theme']},

    python_requires  = '>=3.5',
    install_requires = [
        'urwid>=2.0',
        'urwidtrees>=1.0.3dev0',
        'aiohttp>=3',
        'async_timeout',
        'pyxdg',
        'blinker',
        'natsort',
    ],
    extras_require = {
        'geoip': ['maxminddb'],
        'setproctitle': ['setproctitle'],
    },
    tests_require = [
        'asynctest>=0.11',
    ],

    entry_points = { 'console_scripts': [ 'stig = stig:run' ] },

    classifiers = [
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Environment :: Console',
        'Operating System :: Unix',
        'Development Status :: 3 - Alpha',
    ],
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Documentation commands"""

from ...logging import make_logger
log = make_logger(__name__)


from ..base import misc as base
from .. import (ExpectedResource, CmdError)
from ._common import make_tab_title_widget


class HelpCmd(base.HelpCmdbase):
    provides = {'tui'}
    tui = ExpectedResource

    def display_help(self, topics, lines):
        import urwid
        from ...tui.scroll import ScrollBar
        from ...tui.views.text import SearchableText
        from ...tui.main import keymap

        if hasattr(self, 'title'):
            titlew = make_tab_title_widget(str(self.title),
                                           attr_unfocused='tabs.help.unfocused',
                                           attr_focused='tabs.help.focused')
        else:
            titlew = make_tab_title_widget(','.join(topics),
                                           attr_unfocused='tabs.help.unfocused',
                                           attr_focused='tabs.help.focused')

        helptext_widget_cls = keymap.wrap(SearchableText, context='helptext')
        helptext_widget = helptext_widget_cls(lines)
        textw = urwid.AttrMap(helptext_widget, 'helptext')
        contentw = urwid.AttrMap(ScrollBar(textw), 'scrollbar')
        self.tui.tabs.load(titlew, contentw)


class VersionCmd(base.VersionCmdbase):
    provides = {'tui'}


class LogCmd(base.LogCmdbase):
    provides = {'tui'}
    tui = ExpectedResource

    def _do(self, action, *args):
        logwidget = self.tui.logwidget
        if action == 'clear':
            if len(tuple(logwidget.entries)) < 1:
                raise CmdError()
            else:
                logwidget.clear()

        elif action == 'scroll':
            args = ' '.join(args)
            if args == 'up':
                logwidget.scroll_relative('up', 1)
            elif args == 'down':
                logwidget.scroll_relative('down', 1)
            elif args == 'page up':
                logwidget.scroll_relative('up', logwidget.height-1)
            elif args == 'page down':
                logwidget.scroll_relative('down', logwidget.height-1)
            elif args == 'top':
                logwidget.scroll_to('top')
            elif args == 'bottom':
                logwidget.scroll_to('bottom')
            else:
                raise CmdError('Invalid arguments for "scroll": %r' % (args,))

        else:
            cmd_str = '%s %s' % (action, ' '.join(args))
            raise CmdError('Unsupported command in TUI mode: %s' % cmd_str)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Low-level communication with the Transmission daemon"""

from ...logging import make_logger
log = make_logger(__name__)

import asyncio
import json
import textwrap
from blinker import Signal
import warnings
import async_timeout

from ..errors import (ConnectionError, TimeoutError, RPCError, AuthError, ClientError)


AUTH_ERROR_CODE = 401
CSRF_ERROR_CODE = 409
CSRF_HEADER = 'X-Transmission-Session-Id'
TIMEOUT = 10


class TransmissionRPC():
    """
    Low-level AsyncIO Transmission RPC communication

    This class handles connecting to a Transmission daemon via the RPC
    interface.  It does not implement the RPC protocol, only basic things like
    authentication, sending requests and receiving responses.  High-level RPCs
    are done in the *API classes.
    """

    def __init__(self, host='localhost', port=9091, *, tls=False, user='',
                 password='', path='/transmission/rpc', enabled=True, loop=None):
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self._host = host
        self._port = port
        self._path = path
        self._tls = tls
        self._user = user
        self._password = password
        self._headers = {'content-type': 'application/json'}
        self._session = None
        self._enabled_event = asyncio.Event(loop=loop)
        self.enabled = enabled
        self._request_lock = asyncio.Lock(loop=loop)
        self._connecting_lock = asyncio.Lock(loop=loop)
        self._connection_tested = False
        self._connection_exception = None
        self._timeout = TIMEOUT
        self._version = None
        self._rpcversion = None
        self._rpcversionmin = None
        self._on_connecting = Signal()
        self._on_connected = Signal()
        self._on_disconnected = Signal()
        self._on_error = Signal()

    def __del__(self, _warnings=warnings):
        if self._session is not None and not self._session.closed:
            _warnings.warn('disconnect() wasn\'t called', ResourceWarning)
            asyncio.ensure_future(self._session.close())

    def on(self, signal, callback, autoremove=True):
        """
        Register `callback` for `signal`

        signal: 'connecting', 'connected', 'disconnected' or 'error'
        callback: a callable that receives this instance as a positional
                  argument and, in case of the 'error' signal, the exception as
                  a keyword argument with the name 'error'

        Callbacks are automatically unsubscribed when they are
        garbage-collected.
        """
        try:
            sig = getattr(self, '_on_' + signal)
        except AttributeError:
            raise ValueError('Unknown signal: {!r}'.format(signal))
        else:
            if not isinstance(sig, Signal):
                raise ValueError('Unknown signal: {!r}'.format(signal))
            else:
                log.debug('Registering %r for %r event', callback, signal)
                sig.connect(callback, weak=autoremove)

    @property
    def version(self):
        """Version of the Transmission daemon or None if not connected"""
        return self._version

    @property
    def rpcversion(self):
        """RPC version of the Transmission daemon or None if not connected"""
        return self._rpcversion

    @property
    def rpcversionmin(self):
        """Oldest RPC version supported by Transmission daemon or None if not connected"""
        return self._rpcversionmin

    @property
    def host(self):
        """
        Hostname or IP of the Transmission RPC interface

        Setting this property calls disconnect().
        """
        return self._host
    @host.setter
    def host(self, host):
        self._host = str(host)
        asyncio.ensure_future(self.disconnect('Changing host: %r' % self._host))

    @property
    def path(self):
        """
        Path of the Transmission RPC interface

        Setting this property calls disconnect().
        """
        return self._path
    @path.setter
    def path(self, path):
        if not path or path[0] != '/':
            path = '/' + path
        self._path = path
        asyncio.ensure_future(self.disconnect('Changing path: %r' % self._path))

    @property
    def port(self):
        """
        Port of the Transmission RPC interface

        Setting this property calls disconnect().
        """
        return self._port
    @port.setter
    def port(self, port):
        self._port = int(port)
        asyncio.ensure_future(self.disconnect('Changing port: %r' % self._port))

    @property
    def user(self):
        """
        Username for authenticating to the Transmission RPC interface or empty string

        Setting this property calls disconnect().
        """
        return self._user
    @user.setter
    def user(self, user):
        self._user = str(user)
        asyncio.ensure_future(self.disconnect('Changing user: %r' % self._user))

    @property
    def password(self):
        """
        Password for authenticating to the Transmission RPC interface or empty string

        Setting this property calls disconnect().
        """
        return self._password
    @password.setter
    def password(self, password):
        self._password = str(password)
        asyncio.ensure_future(self.disconnect('Changing password: %r' % self._password))

    @property
    def tls(self):
        """
        Whether to use HTTPS for connecting to the Transmission RPC interface

        Setting this property calls disconnect().
        """
        return self._tls
    @tls.setter
    def tls(self, tls):
        self._tls = bool(tls)
        asyncio.ensure_future(self.disconnect('Changing tls: %r' % self._tls))

    @property
    def url(self):
        """Schema, host, port, and path combined as a string"""
        return '%s://%s:%d%s' % (
            'https' if self.tls else 'http', self.host, self.port, self.path)

    @property
    def timeout(self):
        """Number of seconds to try to connect before giving up"""
        return self._timeout
    @timeout.setter
    def timeout(self, timeout):
        self._timeout = float(timeout)

    @property
    def enabled(self):
        """
        Whether requests should connect

        If this is set to False, requests will wait for it to be set to True.
        This allows you to block any connection attempts until the connection
        parameters (host, user, password, etc) are specified to prevent any
        unwarranted error messages.
        """
        return self._enabled_event.is_set()
    @enabled.setter
    def enabled(self, enabled):
        if enabled and not self.enabled:
            log.debug('Enabling %r', self)
            self._enabled_event.set()
        elif not enabled and self.enabled:
            log.debug('Disabling %r', self)
            self._enabled_event.clear()
            if self.connected:
                asyncio.ensure_future(self.disconnect())

    @property
    def connected(self):
        """Return True if connected, False otherwise"""
        return (self._session is not None
                and not self._session.closed
                and self._connection_tested)

    async def connect(self):
        """
        Connect to running daemon

        If the `enabled` property is set to False, this method blocks until
        `enabled` is set to True.

        Raises RPCError, ConnectionError or AuthError.
        """
        log.debug('Connecting to %s (timeout=%ss)', self.url, self.timeout)
        self._on_connecting.send(self)

        if self._connecting_lock.locked():
            if self._connection_exception is not None:
                # The other connect() call failed
                log.debug('Found connection error: %r', self._connection_exception)
                raise self._connection_exception

            log.debug('Connection is already being established - Waiting ...')
            try:
                async with async_timeout.timeout(self.timeout):
                    await self._enabled_event.wait()
            except asyncio.TimeoutError:
                raise TimeoutError(self.timeout, self.url)
            else:
                if self.connected:
                    log.debug('Connection is up: %r', self.url)
                    return

        async with self._connecting_lock:
            log.debug('Acquired connect() lock')

            if self.connected:
                await self.disconnect('Reconnecting')

            # Block until we're enabled
            await self._enabled_event.wait()

            import aiohttp
            session_args = {'loop': self.loop}
            if self.user and self.password:
                session_args['auth'] = aiohttp.BasicAuth(self.user, self.password,
                                                         encoding='utf-8')
            self._session = aiohttp.ClientSession(**session_args)

            # Check if connection works
            log.debug('Testing connection to %s', self.url)
            try:
                test_request = json.dumps({'method':'session-get'})
                info = await self._send_request(test_request)
            except ClientError as e:
                self._connection_exception = e
                log.debug('Caught during connection test: %r', e)
                await self._reset()
                self._on_error.send(self, error=e)
                raise
            else:
                self._version = info['version']
                self._rpcversion = info['rpc-version']
                self._rpcversionmin = info['rpc-version-minimum']
                self._connection_tested = True
                self._connection_exception = None
                log.debug('Connection established: %s', self.url)
                self._on_connected.send(self)

            log.debug('Releasing connect() lock')

    async def disconnect(self, reason=None):
        """
        Disconnect if connected

        reason: Why are we disconnecting? Only used in a debugging message.
        """
        if self.connected:
            await self._reset()
            log.debug('Disconnecting from %s (%s)', self.url,
                      reason if reason is not None else 'for no reason')
            self._on_disconnected.send(self)

    async def _reset(self):
        if self._session is not None:
            await self._session.close()
        self._session = None
        self._version = None
        self._rpcversion = None
        self._rpcversionmin = None
        self._connection_tested = False

    async def _post(self, data):
        async with async_timeout.timeout(self.timeout):
            response = await self._session.post(self.url, data=data, headers=self._headers)

            if response.status == CSRF_ERROR_CODE:
                # Send request again with CSRF header
                self._headers[CSRF_HEADER] = response.headers[CSRF_HEADER]
                log.debug('Setting CSRF header: %s = %s',
                          CSRF_HEADER, response.headers[CSRF_HEADER])
                await response.release()
                return await self._post(data)

            elif response.status == AUTH_ERROR_CODE:
                await response.release()
                log.debug('Authentication failed: %s: user=%r, password=%r',
                          self.url, self.user, self.password)
                raise AuthError(self.url)

            else:
                import aiohttp
                try:
                    answer = await response.json()
                except aiohttp.ClientResponseError:
                    raise RPCError('Server sent malformed JSON: %s' % await response.text())
                else:
                    return answer

    async def _send_request(self, post_data):
        """
        Send RPC POST request to daemon

        post_data: Any valid RPC request as JSON string

        If applicable, returns response['arguments']['torrents'] or
        response['arguments'], otherwise response.

        Raises ClientError.
        """
        import aiohttp
        try:
            answer = await self._post(post_data)

        # CancelledError is raised when we're writing on a "closing
        # transport". Not sure if this happens in the real world, but it happens
        # in the tests for some reason.
        except (aiohttp.ClientError, asyncio.CancelledError) as e:
            log.debug('Caught during POST request: %r', e)
            raise ConnectionError(self.url)

        except asyncio.TimeoutError as e:
            raise TimeoutError(self.timeout, self.url)

        else:
            if answer['result'] != 'success':
                raise RPCError(answer['result'].capitalize())
            else:
                if 'arguments' in answer:
                    if 'torrents' in answer['arguments']:
                        return answer['arguments']['torrents']
                    else:
                        return answer['arguments']
                return answer

    def __getattr__(self, method):
        """
        Return asyncio coroutine that sends RPC request and returns response

        method: Any method from the RPC specs with every '-' replaced with '_'.
                For arguments see the RPC specs.

        Example:
        >>> stats = await client.session_stats()
        >>> torrents = await client.torrent_get(ids=(1,2,3), fields=('status','name'))

        Raises RPCError, ConnectionError, AuthError
        """
        async def request(arguments=None, **kwargs):
            arguments = arguments or {}

            async with self._request_lock:
                if not self.connected:
                    log.debug('Autoconnecting for %r', method)
                    await self.connect()

                arguments.update(**kwargs)
                rpc_request = json.dumps({'method'    : method.replace('_', '-'),
                                          'arguments' : arguments})

                try:
                    return await self._send_request(rpc_request)
                except ClientError as e:
                    log.debug('Caught ClientError in %r request: %r', method, e)

                    # RPCError does not mean host is unreachable, there was just a
                    # misunderstanding, so we're still connected.
                    if not isinstance(e, RPCError) and self.connected:
                        await self.disconnect(str(e))

                    self._on_error.send(self, error=e)
                    raise

        request.__name__ = method
        request.__qualname__ = method
        return request



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)


class GeoIPError(Exception):
    pass


class GeoIPBase():
    GeoIPError = GeoIPError

    @property
    def cachedir(self):
        """Where to cache the downloaded database"""
        return getattr(self, '_cachedir', None)
    @cachedir.setter
    def cachedir(self, cachedir):
        self._cachedir = str(cachedir)

    @property
    def enabled(self):
        """Whether lookup functions always return None"""
        return getattr(self, '_enabled', False)
    @enabled.setter
    def enabled(self, enabled):
        self._enabled = bool(enabled)

    async def load(self, *args, **kwargs):
        pass

    def country_code(self, addr):
        return None


try:
    import maxminddb
except ImportError:
    class GeoIP(GeoIPBase):
        available = False
else:
    import os
    import time
    import asyncio

    class GeoIP(GeoIPBase):
        available = True
        max_age = 60*60*24*30
        timeout = 5
        url = 'https://geolite.maxmind.com/download/geoip/database/GeoLite2-Country.mmdb.gz'

        def __init__(self):
            super().__init__()
            self.filename = self.url.split('/')[-1]
            if self.filename[-3:] == '.gz':
                self.filename = self.filename[:-3]
            self._lookup_cache = {}

        @property
        def cachefile(self):
            return os.path.join(self.cachedir, self.filename)

        async def load(self, force_update=False, loop=None):
            if loop is None:
                loop = asyncio.get_event_loop()

            # Maybe get fresh database from URL
            if force_update:
                log.debug('Forcing database update')
                await self._update_dbfile(loop)
            else:
                cachefile = self.cachefile
                if os.path.exists(cachefile):
                    age = time.time() - os.path.getmtime(cachefile)
                    if age > self.max_age:
                        log.debug('Cached database is older than %r seconds', self.max_age)
                        await self._update_dbfile(loop)
                else:
                    log.debug('No database found: %r', self.cachefile)
                    await self._update_dbfile(loop)

            # Read database
            try:
                self._db = maxminddb.open_database(cachefile)
            except maxminddb.InvalidDatabaseError as e:
                raise GeoIPError('Unable to read geolocation database %s: Invalid format'
                                 % (cachefile,))
            except FileNotFoundError as e:
                raise GeoIPError('Unable to read geolocation database %s'
                                 % (cachefile,))

        async def _update_dbfile(self, loop):
            import aiohttp
            import gzip

            log.debug('Fetching fresh geolocation database from %s', self.url)

            async with aiohttp.ClientSession() as session:
                try:
                    async with session.get(self.url, timeout=self.timeout) as resp:
                        data = await resp.content.read()
                except aiohttp.ClientError as e:
                    raise GeoIPError('Failed to download geolocation database %s: %s'
                                     % (self.url, os.strerror(e.errno)))

            try:
                data = gzip.decompress(data)
            except OSError:
                pass   # Maybe downloaded data isn't gzipped

            cachedir = self.cachedir
            cachefile = self.cachefile
            try:
                if not os.path.exists(cachedir):
                    os.makedirs(cachedir)
                with open(cachefile, 'wb') as f:
                    f.write(data)
            except OSError as e:
                if os.path.exists(cachefile):
                    os.remove(cachefile)
                raise GeoIPError('Unable to write geolocation database %s: %s'
                                 % (cachefile, os.strerror(e.errno)))

            log.debug('Wrote new geoip database: %s', self.cachefile)

        def country_code(self, addr):
            if not self.enabled:
                return None

            cache = self._lookup_cache
            country, timestamp = cache.get(addr, (None, 0))
            now = time.time()
            if country is not None and now - timestamp < self.max_age:
                return country

            db = getattr(self, '_db', None)
            if db is not None:
                try:
                    info = db.get(addr)
                except maxminddb.InvalidDatabaseError:
                    pass
                else:
                    if isinstance(info, dict):
                        country = info.get('country')
                        if isinstance(country, dict):
                            iso_code = country.get('iso_code')
                            cache[addr] = (iso_code, now)
                            return iso_code



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

import blinker
from collections import abc
from types import SimpleNamespace
import os

from ..poll import RequestPoller
from ..utils import (convert, const, Bool, Option, Int, Path, BoolOrPath, BoolOrBandwidth)


# Transform key (as in `settings[key]`) to property name and vice versa
def _key2property(key):
    return key.replace('.', '_')
def _property2key(property_name):
    return property_name.replace('_', '.')


_SETTINGS = {}
def _setting(value_cls, **kwargs):
    """Decorator for SettingsAPI properties"""
    def wrap(method):
        property_name = method.__name__
        setting_name = _property2key(property_name)
        description = kwargs.pop('description') if 'description' in kwargs else None
        converter = value_cls.partial(**kwargs)

        def get_typespec(api):
            return SimpleNamespace(
                converter=converter,
                description=description or getattr(type(api), property_name).__doc__)
        _SETTINGS[setting_name] = get_typespec

        return property(method)
    return wrap


class SettingsAPI(abc.Mapping, RequestPoller):
    """
    Transmission daemon settings

    `set_*` methods are coroutine functions that request value changes from the
    server.

    `get_*` methods are coroutine functions that fetch values from the server.

    Cached values, which are updated every `interval` seconds, are available as
    properties with 'get_' removed from the equivalent method name
    (e.g. api.get_port() -> api.port).

    Use `on_change` to set a callback for interval updates.

    Cached values are also available as mapping items with '.' instead of '_',
    e.g.  `settings['path.incomplete']`.

    To update cached values, use `update` (async) or `poll` (sync).
    """

    # Mapping methods
    def __getitem__(self, key):
        try:
            item = getattr(self, _key2property(key))
        except AttributeError as e:
            raise KeyError(key)
        else:
            return item

    def __contains__(self, key):
        return key in self._cache

    def __iter__(self):
        return iter(self._cache)

    def __len__(self):
        return len(self._cache)


    def __init__(self, srvapi, interval=1):
        self._cache = {}
        self._descriptions = {}
        self._converters = {}
        for setting,get_typespec in _SETTINGS.items():
            typespec = get_typespec(self)
            self._cache[setting] = None
            self._converters[setting] = typespec.converter
            self._descriptions[setting] = typespec.description

        self._raw = None    # Raw dict from 'session-get' or None if not connected
        self._srvapi = srvapi
        self._on_update = blinker.Signal()

        super().__init__(self._srvapi.rpc.session_get, interval=interval, loop=srvapi.loop)
        self.on_response(self._handle_session_get)
        self.on_error(self._handle_error)

    def description(self, name):
        """Return setting's description"""
        return self._descriptions[name]

    def syntax(self, name):
        """Return setting's value syntax"""
        return self._converters[name].syntax

    def default(self, name):
        # Maintain consistency with local Settings class
        return const.DISCONNECTED

    async def update(self):
        """Request update from server"""
        log.debug('Requesting immediate settings update')
        self._handle_session_get(await self.request())

    def _handle_session_get(self, response):
        """Request update from server"""
        log.debug('Handling settings update')
        self.clearcache(run_callbacks=False)
        self._raw = response
        self._on_update.send(self)

    def _handle_error(self, error):
        self.clearcache(run_callbacks=True)

    def clearcache(self, run_callbacks=True):
        """
        Clear cached settings

        update: Whether to run `on_update` callbacks
        """
        log.debug('Clearing %s cache', type(self).__name__)
        self._raw = None
        for setting in self._cache:
            self._cache[setting] = None
        if run_callbacks:
            self._on_update.send(self)

    def on_update(self, callback, autoremove=True):
        """
        Register `callback` to be called when settings have changed

        `callback` gets the instance of this class.

        If `autoremove` is True, `callback` is removed automatically when it is
        deleted.
        """
        log.debug('Registering %r to receive settings updates', callback)
        self._on_update.connect(callback, weak=autoremove)

    async def get(self, key):
        """Same as __getitem__ but refresh cache first"""
        if key not in self:
            raise ValueError(key)
        else:
            method = getattr(self, 'get_' + _key2property(key))
            return await method()

    async def set(self, key, value):
        """Asynchronous replacement for __setitem__"""
        if key not in self:
            raise ValueError(key)
        else:
            method = getattr(self, 'set_' + _key2property(key))
            await method(value)

    def _get(self, key, field_or_callable):
        """Get setting from cache if possible"""
        if self._cache[key] is None:
            if self._raw is None:
                self._cache[key] = const.DISCONNECTED
            elif callable(field_or_callable):
                self._cache[key] = self._converters[key](field_or_callable())
            else:
                self._cache[key] = self._converters[key](self._raw[field_or_callable])
        return self._cache[key]

    async def _set(self, request):
        """Send 'session-set' request with dictionary `request` and call `update`"""
        log.debug('Sending session-set request: %r', request)
        await self._srvapi.rpc.session_set(request)
        await self.update()


    @_setting(Bool)
    def autostart(self):
        """Whether added torrents are started automatically"""
        return self._get('autostart', 'start-added-torrents')

    async def get_autostart(self):
        """Refresh cache and return `autostart`"""
        await self.update()
        return self.autostart

    async def set_autostart(self, enabled):
        """See `autostart`"""
        value = self._converters['autostart'](enabled)
        await self._set({'start-added-torrents': bool(value)})


    # Network settings

    @_setting(Int, min=0, max=65535, prefix='none')
    def port(self):
        """Port used to communicate with peers"""
        return self._get('port', 'peer-port')

    async def get_port(self):
        """Refresh cache and return `port`"""
        await self.update()
        return self.port

    async def set_port(self, port):
        """See `port`"""
        value = self._converters['port'](port)
        await self._set({'peer-port': value,
                         'peer-port-random-on-start': False})


    @_setting(Bool)
    def port_random(self):
        """Whether to pick a random port when the daemon starts"""
        return self._get('port.random', 'peer-port-random-on-start')

    async def get_port_random(self):
        """Refresh cache and return `port_random`"""
        await self.update()
        return self.port_random

    async def set_port_random(self, port_random):
        """See `port_random`"""
        value = self._converters['port.random'](port_random)
        await self._set({'peer-port-random-on-start': bool(value)})


    @_setting(Bool)
    def port_forwarding(self):
        """Whether to autoconfigure port-forwarding via UPnP/NAT-PMP"""
        return self._get('port.forwarding', 'port-forwarding-enabled')

    async def get_port_forwarding(self):
        """Refresh cache and return `port_forwarding`"""
        await self.update()
        return self.port_forwarding

    async def set_port_forwarding(self, enabled):
        """See `port_forwarding`"""
        value = self._converters['port.forwarding'](enabled)
        await self._set({'port-forwarding-enabled': bool(value)})


    @_setting(Int, min=1, max=65535)
    def limit_peers_global(self):
        """Maximum number of connections for all torrents combined"""
        return self._get('limit.peers.global', 'peer-limit-global')

    async def get_limit_peers_global(self):
        """Refresh cache and return `limit_peers_global`"""
        await self.update()
        return self.limit_peers_global

    async def set_limit_peers_global(self, limit):
        """See `limit_peers_global`"""
        value = self._converters['limit.peers.global'](limit)
        await self._set({'peer-limit-global': value})


    @_setting(Int, min=1, max=65535)
    def limit_peers_torrent(self):
        """Maximum number of connections per torrent"""
        return self._get('limit.peers.torrent', 'peer-limit-per-torrent')

    async def get_limit_peers_torrent(self):
        """Refresh cache and return `limit_peers_torrent`"""
        await self.update()
        return self.limit_peers_torrent

    async def set_limit_peers_torrent(self, limit):
        """See `limit_peers_torrent`"""
        value = self._converters['limit.peers.torrent'](limit)
        await self._set({'peer-limit-per-torrent': value})


    @_setting(Option, options=('required', 'preferred', 'tolerated'),
              description='Protocol encryption policy; "required", "preferred" or "tolerated"')
    def encryption(self):
        """
        Whether protocol encryption is used to mask BitTorrent traffic

        One of the strings 'required', 'preferred' or 'tolerated'.
        """
        return self._get('encryption', 'encryption')

    async def get_encryption(self):
        """Refresh cache and return `encryption`"""
        await self.update()
        return self.encryption

    async def set_encryption(self, encryption):
        """See `encryption`"""
        value = self._converters['encryption'](encryption)
        await self._set({'encryption': value})


    @_setting(Bool)
    def utp(self):
        """Whether to use µTP to mitigate latency issues"""
        return self._get('utp', 'utp-enabled')

    async def get_utp(self):
        """Refresh cache and return `utp`"""
        await self.update()
        return self.utp

    async def set_utp(self, enabled):
        """See `utp`"""
        value = self._converters['utp'](enabled)
        await self._set({'utp-enabled': bool(value)})


    @_setting(Bool)
    def dht(self):
        """Whether to use DHT to discover peers for public torrents"""
        return self._get('dht', 'dht-enabled')

    async def get_dht(self):
        """Refresh cache and return `dht`"""
        await self.update()
        return self.dht

    async def set_dht(self, enabled):
        """See `dht`"""
        value = self._converters['dht'](enabled)
        await self._set({'dht-enabled': bool(value)})


    @_setting(Bool)
    def pex(self):
        """Whether to use PEX to discover peers for public torrents"""
        return self._get('pex', 'pex-enabled')

    async def get_pex(self):
        """Refresh cache and return `pex`"""
        await self.update()
        return self.pex

    async def set_pex(self, enabled):
        """See `pex`"""
        value = self._converters['pex'](enabled)
        await self._set({'pex-enabled': bool(value)})


    @_setting(Bool)
    def lpd(self):
        """Whether to use LPD to discover peers for public torrents"""
        return self._get('lpd', 'lpd-enabled')

    async def get_lpd(self):
        """Refresh cache and return `lpd`"""
        await self.update()
        return self.lpd

    async def set_lpd(self, enabled):
        """See `lpd`"""
        value = self._converters['lpd'](enabled)
        await self._set({'lpd-enabled': bool(value)})


    # Local Filesystem settings

    @_setting(Path)
    def path_complete(self):
        """Where to put downloaded files"""
        return self._get('path.complete', 'download-dir')

    async def get_path_complete(self):
        """Refresh cache and return `path_complete`"""
        await self.update()
        return self.path_complete

    async def set_path_complete(self, path):
        """
        Set download directory files

        If path is relative (i.e. doesn't start with a path separator
        (e.g. '/')), it is relative to what `get_path_complete` returns.
        """
        value = self._converters['path.complete'](path)
        if not value.startswith(os.sep):
            current_path = await self.get_path_complete()
            value = os.path.join(current_path, value)
        await self._set({'download-dir': value})


    @_setting(BoolOrPath,
              description='Where to put partially downloaded files')
    def path_incomplete(self):
        """
        Path to incomplete files or Bool(<False>) to put them in `path_complete`
        """
        def get_raw_value():
            if self._raw['incomplete-dir-enabled']:
                return self._raw['incomplete-dir']
            else:
                return False
        return self._get('path.incomplete', get_raw_value)

    async def get_path_incomplete(self):
        """Refresh cache and return `path_incomplete`"""
        await self.update()
        return self.path_incomplete

    async def set_path_incomplete(self, path):
        """
        Set path to incomplete files or Bool(<False>)

        If `path` is not a `str` instance, it is evaluated as a bool and this
        feature is enabled or disabled accordingly without changing the path.

        If path is relative (i.e. doesn't start with a path separator
        (e.g. '/')), it is relative to what `get_path_incomplete` returns.
        """
        value = self._converters['path.incomplete'](path)
        request = {'incomplete-dir-enabled': bool(value)}
        if isinstance(value, Path):
            if not value.startswith(os.sep):
                current_path = await self.get_path_incomplete()
                value = os.path.join(current_path, value)
            request['incomplete-dir'] = str(value)
        await self._set(request)


    @_setting(Bool)
    def files_part(self):
        """Whether ".part" is appended to incomplete files"""
        return self._get('files.part', 'rename-partial-files')

    async def get_files_part(self):
        """Refresh cache and return `files_part`"""
        await self.update()
        return self.files_part

    async def set_files_part(self, enabled):
        """See `files_part`"""
        value = self._converters['files.part'](enabled)
        await self._set({'rename-partial-files': bool(value)})


    # Rate limits

    @staticmethod
    def _limit_rate_fields(direction, alt):
        if alt:
            field_value = 'alt-speed-'+direction
            field_enabled = 'alt-speed-enabled'
        else:
            field_value = 'speed-limit-'+direction
            field_enabled = 'speed-limit-'+direction+'-enabled'
        return (field_value, field_enabled)

    @staticmethod
    def _limit_rate_key(direction, alt):
        if alt:
            return 'limit.rate.alt.%s' % direction
        else:
            return 'limit.rate.%s' % direction

    def _get_limit_rate(self, direction, alt=False):
        key = self._limit_rate_key(direction, alt)
        if self._cache[key] is None:
            field_value, field_enabled = self._limit_rate_fields(direction, alt)
            if self._raw is None:
                self._cache[key] = const.DISCONNECTED
            elif self._raw[field_enabled]:
                # Transmission reports kilobytes
                self._cache[key] = self._converters[key](
                    convert.bandwidth(self._raw[field_value]*1000, unit='B'))
            else:
                self._cache[key] = const.UNLIMITED
        return self._cache[key]

    async def _set_limit_rate(self, limit, direction, alt=False):
        key = self._limit_rate_key(direction, alt)
        field_value, field_enabled = self._limit_rate_fields(direction, alt)
        limit = self._converters[key](limit)
        if isinstance(limit, Bool):
            await self._set({field_enabled: bool(limit)})
        elif 0 <= limit < float('inf'):
            raw_limit = round(round(limit.copy(convert_to='B')) / 1000)
            await self._set({field_enabled: True,
                             field_value: raw_limit})
        else:
            await self._set({field_enabled: False})

    async def _adjust_limit_rate(self, adjustment, direction, alt=False):
        key = self._limit_rate_key(direction, alt)
        field_value, field_enabled = self._limit_rate_fields(direction, alt)
        prop_name = _key2property(key)

        current_limit = getattr(self, prop_name)
        if current_limit is const.DISCONNECTED:
            current_limit = await getattr(self, 'get_' + prop_name)()
        adjustment = self._converters[key](adjustment)
        new_limit = BoolOrBandwidth.adjust(current_limit, adjustment)
        await self._set_limit_rate(new_limit, direction, alt=alt)

    @_setting(BoolOrBandwidth,
              description='Global upload rate limit')
    def limit_rate_up(self):
        """
        Upload rate limit

        This uses the application-wide `stig.utils.convert` module to get
        consistent units and unit prefixes.
        """
        return self._get_limit_rate('up', alt=False)

    async def get_limit_rate_up(self):
        """Refresh cache and return `limit_rate_up`"""
        await self.update()
        return self.limit_rate_up

    async def set_limit_rate_up(self, limit):
        """
        Set upload rate limit to `limit` (see also `limit_rate_up`)

        An existing limit is disabled if `limit` is False, the constant
        UNLIMITED, or a number < 0.

        A previously set and then disabled limit is enabled if `limit` is True.
        """
        await self._set_limit_rate(limit, 'up', alt=False)

    async def adjust_limit_rate_up(self, limit):
        """Adjust current upload rate limit by `limit` (positive or negative number)"""
        await self._adjust_limit_rate(limit, 'up', alt=False)


    @_setting(BoolOrBandwidth,
              description='Global download rate limit')
    def limit_rate_down(self):
        """Download rate limit (see `limit_rate_up`)"""
        return self._get_limit_rate('down', alt=False)

    async def get_limit_rate_down(self):
        """Refresh cache and return `limit_rate_down`"""
        await self.update()
        return self.limit_rate_down

    async def set_limit_rate_down(self, limit):
        """Set download rate limit to `limit` (see `set_limit_rate_up`)"""
        await self._set_limit_rate(limit, 'down', alt=False)

    async def adjust_limit_rate_down(self, limit):
        """Adjust current download rate limit by `limit` (positive or negative number)"""
        await self._adjust_limit_rate(limit, 'down', alt=False)


    @_setting(BoolOrBandwidth,
              description='Alternative global upload rate limit')
    def limit_rate_alt_up(self):
        """Alternative upload rate limit (see `limit_rate_up`)"""
        return self._get_limit_rate('up', alt=True)

    async def get_limit_rate_alt_up(self):
        """Refresh cache and return `limit_rate_alt_up`"""
        await self.update()
        return self.limit_rate_alt_up

    async def set_limit_rate_alt_up(self, limit):
        """Set alternative upload rate limit to `limit` (see `set_limit_rate_up`)"""
        await self._set_limit_rate(limit, 'up', alt=True)

    async def adjust_limit_rate_alt_up(self, limit):
        """Adjust current alternative upload rate limit by `limit`"""
        await self._adjust_limit_rate(limit, 'up', alt=True)


    @_setting(BoolOrBandwidth,
              description='Alternative global download rate limit')
    def limit_rate_alt_down(self):
        """Alternative download rate limit (see `limit_rate_up`)"""
        return self._get_limit_rate('down', alt=True)

    async def get_limit_rate_alt_down(self):
        """Refresh cache and return `limit_rate_alt_down`"""
        await self.update()
        return self.limit_rate_alt_down

    async def set_limit_rate_alt_down(self, limit):
        """Set alternative upload rate limit to `limit` (see `set_limit_rate_up`)"""
        await self._set_limit_rate(limit, 'down', alt=True)

    async def adjust_limit_rate_alt_down(self, limit):
        """Adjust current alternative download rate limit by `limit`"""
        await self._adjust_limit_rate(limit, 'down', alt=True)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

import urwid
from ...utils.string import strwidth
from ..scroll import Scrollable


class SearchableText(urwid.WidgetWrap):
    palette_name = 'find.highlight'

    def __init__(self, lines):
        self._lines = lines
        self._search_phrase = None
        self._match_indexes = None
        self._render_action = None
        super().__init__(self._make_content([urwid.Text(line) for line in lines]))

    def _make_content(self, text_widgets):
         return Scrollable(urwid.Pile(text_widgets))

    @property
    def original_widget(self):
        # original_widget is only needed for ScrollBar
        return self._w

    @property
    def search_phrase(self):
        return self._search_phrase

    @search_phrase.setter
    def search_phrase(self, phrase):
        phrase = str(phrase)
        prev_search_phrase = self._search_phrase
        self._match_indexes = None
        if not phrase and prev_search_phrase:
            self._search_phrase = None
            self._case_sensitive = None
            self._clear_matches()
        elif phrase != prev_search_phrase:
            # Case-insensitive matching if phrase is equal to casefolded phrase
            phrase_cf = phrase.casefold()
            case_sensitive = phrase != phrase_cf
            if not case_sensitive:
                # Do case-insensitive matching
                phrase = phrase_cf
            self._search_phrase = phrase
            self._case_sensitive = case_sensitive
            self._highlight_matches()

    def _clear_matches(self):
        scrollpos = self._w.get_scrollpos()
        self._w = self._make_content([urwid.Text(line) for line in self._lines])
        self._w.set_scrollpos(scrollpos)

    def _highlight_matches(self):
        scrollpos = self._w.get_scrollpos()
        phrase = self._search_phrase
        case_sensitive = self._case_sensitive

        def match_boundaries(line):
            """Yield (start, stop) tuple for each occurence of `phrase` in `line`"""
            start = stop = -1
            if not phrase:
                yield (start, stop)

            def positions_of(phrase, line):
                start = line.find(phrase)
                while start > -1:
                    stop = start + strwidth(phrase)
                    yield (start, stop)

                    # Find next start
                    offset = line[stop:].find(phrase)
                    if offset > -1:
                        start = stop + offset
                    else:
                        start = -1

            if case_sensitive:
                yield from positions_of(phrase, line)
            else:
                yield from positions_of(phrase, line.casefold())

        texts = []
        palette_name = self.palette_name
        for line_index,line in enumerate(self._lines):
            line_parts = []

            # Highlight all matches in the line
            prev_hl_stop = 0
            for hl_start,hl_stop in match_boundaries(line):
                before_hl = line[prev_hl_stop:hl_start]
                hl = line[hl_start:hl_stop]
                if before_hl: line_parts.append(before_hl)
                if hl: line_parts.append((palette_name, hl))
                prev_hl_stop = hl_stop

            # Append everything after the final match in the line
            line_parts.append(line[prev_hl_stop:])
            texts.append(urwid.Text(line_parts))

        # This calls _invalidate()
        self._w = self._make_content(texts)

        # Restore scrolling position which was lost when content changed
        self._w.set_scrollpos(scrollpos)

    _JUMP_TO_NEXT_MATCH = object()
    _JUMP_TO_PREV_MATCH = object()
    _MAYBE_JUMP_TO_NEXT_MATCH = object()
    _MAYBE_JUMP_TO_PREV_MATCH = object()

    def _assert_search_phrase_not_None(self):
        if self._search_phrase is None:
            raise RuntimeError("Can't jump to next match with search_phrase set to None")

    def jump_to_next_match(self):
        self._assert_search_phrase_not_None()
        self._render_action = self._JUMP_TO_NEXT_MATCH
        self._invalidate()

    def jump_to_prev_match(self):
        self._assert_search_phrase_not_None()
        self._render_action = self._JUMP_TO_PREV_MATCH
        self._invalidate()

    def maybe_jump_to_next_match(self):
        """Jump to next match if nothing matches on the current page"""
        if self._search_phrase:
            self._assert_search_phrase_not_None()
            self._render_action = self._MAYBE_JUMP_TO_NEXT_MATCH
            self._invalidate()

    def maybe_jump_to_prev_match(self):
        """Jump to previous match if nothing matches on the current page"""
        if self._search_phrase:
            self._assert_search_phrase_not_None()
            self._render_action = self._MAYBE_JUMP_TO_PREV_MATCH
            self._invalidate()

    def render(self, size, focus=False):
        ra = self._render_action
        if self._search_phrase is not None and ra is not None:
            # Calculate indexes of matching rows on the canvas
            if self._match_indexes is None:
                self._find_match_indexes(size)

            # Move next/previous match to the top of the canvas
            if ra is self._JUMP_TO_NEXT_MATCH:
                self._jump_to_next_match(size)
            elif ra is self._JUMP_TO_PREV_MATCH:
                self._jump_to_prev_match(size)

            # Same as above, but only if there are no matches visible
            elif ra is self._MAYBE_JUMP_TO_NEXT_MATCH:
                if not self._any_matches_on_current_page(size):
                    self._jump_to_next_match(size)
            elif ra is self._MAYBE_JUMP_TO_PREV_MATCH:
                if not self._any_matches_on_current_page(size):
                    self._jump_to_prev_match(size)

            self._render_action = None
        return super().render(size, focus)

    def _jump_to_next_match(self, size):
        scrollpos = self._w.get_scrollpos()
        for i in self._match_indexes:
            if i > scrollpos:
                self._w.set_scrollpos(i)
                break

    def _jump_to_prev_match(self, size):
        scrollpos = self._w.get_scrollpos()
        for i in reversed(self._match_indexes):
            if i < scrollpos:
                self._w.set_scrollpos(i)
                break

    def _any_matches_on_current_page(self, size):
        viewport_min = self._w.get_scrollpos()
        viewport_max = viewport_min + size[1] - 1
        for i in self._match_indexes:
            if viewport_min <= i <= viewport_max:
                return True
        return False

    def _find_match_indexes(self, size):
        phrase = self._search_phrase
        case_sensitive = self._case_sensitive

        def row_matches(row):
            for _,_,text in row:
                text_dec = text.decode()
                if case_sensitive and phrase in text_dec:
                    return True
                elif not case_sensitive and phrase in text_dec.casefold():
                    return True
            return False

        # Render the full Pile canvas because long lines can cause line breaks
        # and so we can't just get indexes from self._lines.
        full_canv = self._w.base_widget.render((size[0],), False)
        indexes = []
        for i,row in enumerate(full_canv.content()):
            text = ''.join(p[2].decode() for p in row).strip()
            if row_matches(row):
                indexes.append(i)
        self._match_indexes = indexes



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

import urwid
import urwidtrees
from collections import abc
import builtins

from .file import TUICOLUMNS
from . import (ItemWidgetBase, ListWidgetBase, stringify_torrent_filter)
from ...client import TorrentFileFilter


from ...views.file import TorrentFileDirectory
from urwidtrees.decoration import ArrowTree
class FileTreeDecorator(ArrowTree):
    """urwidtrees decorator for TorrentFiles and TorrentFileTrees"""

    def __init__(self, torrents, keymap, table, ffilter):
        self._filewidgetcls = keymap.wrap(FileItemWidget, context='file')
        self._table = table
        self._ffilter = ffilter
        self._widgets = {}
        self._filtered_counts = {}
        forest = self._create_file_forest(torrents)
        super().__init__(forest, indent=2)

    def _file_is_filtered(self, tfile):
        if self._ffilter is None:
            return False  # No filter specified
        elif isinstance(self._ffilter, (abc.Sequence, abc.Set)):
            # ffilter is a collection of file IDs
            return not tfile['id'] in self._ffilter
        else:
            # ffilter is a TorrentFileFilter instance
            return not self._ffilter.match(tfile)

    def _create_file_forest(self, torrents):
        # Create a list of nested trees in SimpleTree format.  But the leaves
        # are mappings instead of widgets.  Each mapping contains the
        # information that the `decorate` method needs to create a widget.

        def create_tree(node, content):
            if content.nodetype == 'leaf':
                # Process single file
                if not self._file_is_filtered(content):
                    return (content, None)
                else:
                    return None

            elif content.nodetype == 'parent':
                # Recurse into directory entries (files and subdirectories)
                tree = []
                filtered_count = 0
                for k,v in sorted(content.items(), key=lambda pair: pair[0].lower()):
                    if v.nodetype == 'leaf':
                        if not self._file_is_filtered(v):
                            tree.append((v, None))
                            nonlocal fcount
                            fcount += 1
                        else:
                            filtered_count += 1
                    elif v.nodetype == 'parent':
                        dirnode = TorrentFileDirectory(name=k, tree=v)
                        tree.append(create_tree(dirnode, v))

                self._filtered_counts[node['id']] = filtered_count
                node['name'] = TorrentFileDirectory.create_directory_name(node['name'], filtered_count)
                return (node, tree or None)

        fcount = 0
        forest = []  # Multiple trees as siblings
        for t in sorted(torrents, key=lambda t: t['name'].lower()):
            filetree = t['files']
            filecount = len(tuple(filetree.files))
            if filecount > 0:
                nodetree = None
                rootnodename = next(iter(filetree.keys()))
                if filecount == 1:
                    # Single-file torrent
                    if not self._file_is_filtered(filetree):
                        nodetree = (filetree[rootnodename], None)
                else:
                    # Torrent with directory structure
                    # If we don't ignore the topmost node, the torrent itself is
                    # displayed as a directory.
                    content = filetree[rootnodename]
                    rootnode = TorrentFileDirectory(rootnodename, tree=content)
                    nodetree = create_tree(rootnode, content)

                if nodetree is not None:
                    forest.append(nodetree)

        self.filecount = fcount
        return forest

    def decorate(self, pos, data, is_first=True):
        # We can use the tree position as table ID
        self._table.register(pos)
        row = self._table.get_row(pos)

        # We use parent's decorate() method to give the name column a tree
        # structure.  But we also need the original update() method so we can
        # apply new data to the widget.  This is dirty but it works.
        if row.exists('name'):
            update_method = row.name.update
            decowidget = super().decorate(pos, row.name, is_first=is_first)
            decowidget.update = update_method
            row.replace('name', decowidget)

        # Wrap the whole row in a FileItemWidget with keymapping.  This also
        # applies all the other values besides the name (size, progress, etc).
        file_widget = self._filewidgetcls(data, row)
        self._widgets[data['id']] = file_widget
        return file_widget

    def update(self, torrents):
        widgets = self._widgets
        for t in torrents:
            tid = t['id']

            # Update file nodes
            for f in t['files'].files:
                node_id = f['id']
                # File might be filtered
                if node_id in widgets:
                    widgets[node_id].update(f)

            # Update directory nodes
            for name,content in t['files'].directories:
                node_id = content.id
                if node_id in widgets:
                    filtered_count = self._filtered_counts[node_id]
                    data = TorrentFileDirectory(name, tree=content,
                                                filtered_count=filtered_count)
                    widgets[node_id].update(data)

    @property
    def widgets(self):
        """Yield all file and directory widgets in this tree"""
        yield from self._widgets.values()



class FileItemWidget(ItemWidgetBase):
    palette_unfocused = 'filelist'
    palette_focused   = 'filelist.focused'
    columns_focus_map = {}
    for col in TUICOLUMNS.values():
        columns_focus_map.update(col.style.focus_map)

    @property
    def id(self):
        return self.data['id']

    @property
    def torrent_id(self):
        return self.data['tid']

    @property
    def nodetype(self):
        """'parent' or 'leaf'"""
        return self.data.nodetype


class FileListWidget(ListWidgetBase):
    tuicolumns      = TUICOLUMNS
    ListItemClass   = FileItemWidget
    keymap_context  = 'file'
    palette_name    = 'filelist'
    focusable_items = True

    def __init__(self, srvapi, keymap, tfilter, ffilter, columns=None, title=None):
        super().__init__(srvapi, keymap, columns=columns, title=title)
        self._tfilter = tfilter
        self._ffilter = ffilter
        self._secondary_filter = None
        self._initialized = False
        self._torrents = None

        self._poller = self._srvapi.create_poller(
            self._srvapi.torrent.torrents, tfilter, keys=('files', 'name')
        )
        self._poller.on_response(self._handle_files)

    def _handle_files(self, response):
        if response is None or not response.torrents:
            self.clear()
        else:
            if self._initialized:
                self._update_listitems(response.torrents)
            else:
                self._init_listitems(response.torrents)
                self._initialized = True
        self._invalidate()

    def _init_listitems(self, torrents):
        self.clear()
        if torrents:
            self._torrents = torrents
            self._create_filetree()
            self._listbox._invalidate()

    def _create_filetree(self):
        # Combine primary and secondary file filters
        ffilter = self._ffilter
        sffilter = self._secondary_filter
        if ffilter is None:
            ffilter = sffilter
        elif sffilter is not None:
            ffilter = ffilter & sffilter

        self._filetree = FileTreeDecorator(self._torrents, self._keymap, self._table, ffilter)
        self._listbox.body = urwidtrees.widgets.TreeListWalker(self._filetree)

    def _update_listitems(self, torrents=()):
        if torrents:
            self._filetree.update(torrents)
            self._torrents = torrents

    @property
    def secondary_filter(self):
        return self._secondary_filter

    @secondary_filter.setter
    def secondary_filter(self, file_filter):
        if file_filter is None:
            self._secondary_filter = None
        else:
            self._secondary_filter = TorrentFileFilter(file_filter)
        self._create_filetree()


    @property
    def title(self):
        if self._title_name is None:
            if self._torrents is None:
                return ''
            else:
                title = stringify_torrent_filter(self._tfilter, self._torrents)
                if self._ffilter:
                    title += ' %s' % self._ffilter
                return title
        else:
            return super().title

    def clear(self):
        # We can't call super().clear() because it runs the more efficient
        # `self._listbox.body[:] = ()`
        # That doesn't work here because urwidtrees.TreeListWalker doesn't
        # support item assignment.
        self._listbox.body = urwid.SimpleListWalker([])
        self._listbox._invalidate()
        self._initialized = False
        self._table.clear()
        self._marked.clear()

    def refresh(self):
        self._poller.poll()

    @property
    def count(self):
        return self._filetree.filecount if hasattr(self, '_filetree') else 0

    @property
    def focus_position(self):
        positions = tuple(self._filetree.positions())
        return positions.index(self._listbox.focus_position)

    @focus_position.setter
    def focus_position(self, focus_position):
        positions = tuple(self._filetree.positions())
        i = min(focus_position, len(positions)-1)
        try:
            self._listbox.focus_position = positions[i]
        except KeyError:
            pass

    @property
    def focused_file_ids(self):
        """File IDs of the focused files in a tuple"""
        focused = self.focused_widget
        if focused is not None:
            # The focused widget in the list can be a file or a directory.  If
            # it's a directory, the 'id' property returns the IDs of all the
            # contained files recursively, otherwise it's just a single ID.  For
            # consistency, we always return a sequence of IDs.
            if focused.nodetype == 'leaf':
                return (focused.id,)
            else:
                return focused.id

    @property
    def focused_torrent_id(self):
        """Torrent ID of the currently focused file or `None`"""
        focused_widget = self.focused_widget
        if focused_widget is not None:
            return focused_widget.torrent_id


    def all_children(self, pos):
        """Yield (position, widget) tuples of all sub-nodes (leaves and parents)"""
        ft = self._filetree
        lb = self._listbox
        def recurse(subpos):
            widget = lb.body[subpos]
            if ft.is_leaf(subpos):
                yield (subpos, widget)
            else:
                # Yield sub-parent nodes, but not the starting node that was
                # passed to us
                if subpos != pos:
                    yield (subpos, widget)

                new_subpos = ft.first_child_position(subpos)
                while new_subpos is not None:
                    yield from recurse(new_subpos)
                    new_subpos = ft.next_sibling_position(new_subpos)

        yield from recurse(pos)

    def _set_mark(self, mark, toggle=False, all=False):
        if toggle:
            focused = self.focused_widget
            if focused is not None:
                mark = not focused.is_marked

        def get_widget(pos):
            return self._listbox.body[pos]

        def mark_leaves(pos, mark):
            get_widget(pos).is_marked = mark

            for subpos,widget in self.all_children(pos):
                if widget.nodetype == 'leaf':
                    widget.is_marked = mark
                    if mark:
                        self._marked.add(widget)
                    else:
                        self._marked.discard(widget)

                elif widget.nodetype == 'parent':
                    if pos != subpos:  # Avoid infinite recursion
                        mark_leaves(subpos, mark)

        if all:
            # Top ancestor node positions are (0,), (1,), (3,) etc
            for pos in self._filetree.positions():
                if len(pos) == 1:
                    mark_leaves(pos, mark)
        else:
            mark_leaves(self._listbox.focus_position, mark)
        assert builtins.all(m.nodetype == 'leaf' for m in self._marked)

        # A parent node is marked only if all its children are marked.  To check
        # that, we walk through every ancestor up to the top and check all its
        # children.  There is no need to check the children of other parent
        # nodes (uncles, great uncles, etc) because they should already be
        # marked properly from previous runs.

        def all_children_marked(pos):
            marked = True
            childpos = self._filetree.first_child_position(pos)
            while childpos is not None:
                marked = marked and get_widget(childpos).is_marked
                childpos = self._filetree.next_sibling_position(childpos)
            return marked

        parpos = self._filetree.parent_position(self._listbox.focus_position)
        while parpos is not None:
            parwidget = get_widget(parpos)
            parwidget.is_marked = all_children_marked(parpos)
            parpos = self._filetree.parent_position(parpos)

    def refresh_marks(self):
        for widget in self._filetree.widgets:
            widget.is_marked = widget.is_marked



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from ..base import tracker as base
from . import _mixin as mixin
from .. import (ExpectedResource, CmdError)
from ._table import print_table


class ListTrackersCmd(base.ListTrackersCmdbase,
                      mixin.make_request, mixin.select_torrents):
    provides = {'cli'}
    srvapi = ExpectedResource

    async def make_tracker_list(self, torfilter, trkfilter, sort, columns):
        response = await self.make_request(
            self.srvapi.torrent.torrents(torfilter, keys=('name', 'trackers')),
            quiet=True)
        torrents = response.torrents

        if len(torrents) < 1:
            raise CmdError()

        if trkfilter is None:
            filter_trackers = lambda trackers: trackers
        else:
            filter_trackers = lambda trackers: trkfilter.apply(trackers)

        trklist = []
        for torrent in sorted(torrents, key=lambda t: t['name'].lower()):
            trklist.extend(filter_trackers(torrent['trackers']))

        sort.apply(trklist, inplace=True)

        if trklist:
            from ...views.tracker import COLUMNS as TRACKER_COLUMNS
            print_table(trklist, columns, TRACKER_COLUMNS)
        else:
            filter_is_relevant = lambda f: f and str(f) != 'all'

            if filter_is_relevant(trkfilter):
                errmsg = 'No matching trackers'
            else:
                errmsg = 'No trackers'

            if filter_is_relevant(torfilter):
                errmsg += ' in %s torrents' % torfilter

            if filter_is_relevant(trkfilter):
                errmsg += ': %s' % trkfilter

            raise CmdError(errmsg)


class AnnounceCmd(base.AnnounceCmdbase,
                  mixin.make_request, mixin.select_torrents):
    provides = {'cli'}


class TrackerCmd(base.TrackerCmdbase,
                 mixin.make_request, mixin.select_torrents):
    provides = {'cli'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..utils.usertypes import (Bool, Int, Float, Option, Path, multitype)
from ..utils import convert
from . import constants as const


class Bandwidth(Int):
    typename = 'bandwidth'

    def __new__(cls, value, **kwargs):
        value = convert.bandwidth(value)
        kwargs.update(unit=convert.bandwidth.unit,
                      prefix=convert.bandwidth.prefix)
        return super().__new__(cls, value, **kwargs)

    @classmethod
    def _get_syntax(cls, **kwargs):
        return '%s[b|B]' % super()._get_syntax(**kwargs)


class BoolOrBandwidth(multitype(Bool.partial(true=('limited', 'enabled', 'yes', 'on', 'true'),
                                             false=('unlimited', 'disabled', 'no', 'off', 'false')),
                                Bandwidth)):
    @staticmethod
    def adjust(current, adjustment):
        """Adjust `current` by `adjustment`"""
        if isinstance(current, Bool) or current >= float('inf'):
            # If current number is infinity, adjust from 0
            current = 0
        new = current + adjustment
        if new < 0:
            # Drop to 0 if current is greater than zero.
            # If current already is zero, drop to infinity.
            if current > 0:
                new = 0
            else:
                new = const.UNLIMITED
        return new

    def __new__(cls, value, **kwargs):
        if isinstance(value, (float, int)) and value >= float('inf'):
            return super().__new__(cls, 'unlimited')
        else:
            return super().__new__(cls, value, **kwargs)


BoolOrPath = multitype(Bool, Path)


class PerfectInterval():
    """Remove processing time from intervals"""

    def __init__(self, loop):
        self._loop = loop
        self._last_timestamp = 0

    def __call__(self, seconds):
        now = self._loop.time()
        if self._last_timestamp <= 0:
            self._last_timestamp = int(now)
            return seconds
        else:
            expected = self._last_timestamp + seconds
            diff = now - expected
            interval = max(seconds - diff, 0)
            self._last_timestamp = expected
            return interval


import asyncio
from async_timeout import timeout as async_timeout
class SleepUneasy():
    """Asynchronous sleep() that can be aborted"""

    def __init__(self, loop):
        self.loop = loop
        self._interrupt = asyncio.Event(loop=self.loop)
        self._perfint = PerfectInterval(self.loop)

    async def sleep(self, seconds):
        """Sleep for `seconds` or until `interrupt` is called"""
        self._interrupt.clear()
        # Remove processing time from seconds
        seconds = self._perfint(seconds)
        try:
            async with async_timeout(seconds, loop=self.loop):
                await self._interrupt.wait()
        except asyncio.TimeoutError:
            pass  # Interval passed without interrupt
        finally:
            self._interrupt.clear()

    def interrupt(self):
        """Stop sleeping"""
        self._interrupt.set()


from types import SimpleNamespace
class Response(SimpleNamespace):
    """Response to an API call

    All API implementations should use this class to provide return values to
    API calls.

    success: Whether the call was a success
    msgs: Sequence of messages; either strings or ClientError exceptions

    Any other keyword arguments are made available as attributes.
    """
    def __init__(self, success, msgs=(), errors=(), **kwargs):
        super().__init__(success=bool(success),
                         msgs=tuple(msgs), errors=tuple(errors),
                         **kwargs)


def lazy_property(after_creation=None):
    """Property that replaces itself with the requested object when accessed

    `after_creation` is called with the instance of the property.
    """
    # https://stackoverflow.com/a/6849299
    class _lazy_property():
        def __init__(self, fget):
            self.fget = fget
            self.func_name = fget.__name__
            self.after_creation = after_creation

        def __get__(self, obj, cls):
            if obj is None:
                return None
            value = self.fget(obj)
            setattr(obj, self.func_name, value)
            if self.after_creation is not None:
                self.after_creation(obj)
            return value

    return _lazy_property


class LazyDict(dict):
    """Dictionary with callables as values that return the actual value on demand"""
    def __getitem__(self, key):
        value = dict.__getitem__(self, key)
        if callable(value):
            value = value()
            dict.__setitem__(self, key, value)
        return value


import re
class URL():
    """Parse URL as lenient as possible (no validation)"""

    ### Pilfered from yurl: https://github.com/homm/yurl
    # This is not validating regexp.
    # It splits url to unambiguous parts according RFC.
    _split_re = re.compile(r'''
        (?:([^:/?#]+):)?            # scheme
        (?://                       # authority
            (?:([^/?\#@]*)@)?       # userinfo
            ([^/?\#]*)              # host:port
        )?
        ([^?\#]*)                   # path
        \??([^\#]*)                 # query
        \#?(.*)                     # fragment
        ''', re.VERBOSE | re.DOTALL).match

    @classmethod
    def _parse_url(cls, url):
        # Default to http scheme so 'host:123' is not interpreted as 'host://123/'
        if len(url) > 0 and '://' not in str(url):
            url = 'http://' + str(url)

        groups = cls._split_re(url).groups('')
        # Order: scheme, user+password, host+port, path, query, fragment
        dct = {}
        dct['scheme']   = groups[0] or None
        dct['path']     = groups[3] or None
        dct['query']    = groups[4] or None
        dct['fragment'] = groups[5] or None

        auth = groups[1]
        user, pw = None, None
        if auth:
            # Password in authentication is optional
            if ':' in auth:
                user, pw = auth.split(':')
            else:
                user, pw = auth, None
        dct['user'] = user or None
        dct['password'] = pw or None

        # Split port from host only if it consists of digits.
        host = groups[2]
        port = None
        port_idx = host.rfind(':')
        if port_idx >= 0:
            port_str = host[port_idx+1:]
            if port_str.isdigit():
                host = host[:port_idx]
                port = int(port_str)
        dct['host'] = host or None
        dct['port'] = port or None

        return dct

    _obj_cache = {}
    def __new__(cls, url):
        if isinstance(url, cls):
            return url

        cache_id = url
        cache = cls._obj_cache
        url_dict = cache.get(cache_id)
        if url_dict is None:
            url_dict = cache[cache_id] = cls._parse_url(url)

        obj = super().__new__(cls)
        for attr in ('scheme', 'host', 'port', 'path', 'user', 'password'):
            setattr(obj, attr, url_dict[attr])
        return obj

    @property
    def has_auth(self):
        """Whether user and password properties are set"""
        return self.user is not None and self.password is not None

    @property
    def domain(self):
        domain = getattr(self, '_domain_cached', None)
        if domain is None:
            host = self.host
            if not host:
                self._domain_cached = None
            else:
                if host.count('.') <= 1:
                    domain = self._domain_cached = host
                else:
                    parts = host.rsplit('.', maxsplit=2)
                    domain = self._domain_cached = '.'.join(parts[-2:])
        return domain

    def __str__(self):
        string = getattr(self, '_cached_string', None)
        if string is None:
            parts = []
            if self.scheme:
                parts.append('%s://' % self.scheme)
            if self.user:
                parts.append('%s' % self.user)
                if self.password:
                    parts.append(':%s' % self.password)
                parts.append('@')
            if self.host:
                parts.append(self.host)
            if self.port:
                parts.append(':%s' % self.port)
            if self.path:
                parts.append('%s' % self.path)
            string = self._cached_string = ''.join(parts)
        return string

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name[0] != '_':
            # Clear caches
            if hasattr(self, '_cached_string'):
                del self._cached_string
            if name == 'host' and hasattr(self, '_domain_cached'):
                del self._domain_cached

    def __repr__(self):
        return '<%s %s>' % (type(self).__name__, str(self))

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __le__(self, other):
        return str(self) <= str(other)

    def __gt__(self, other):
        return str(self) > str(other)

    def __ge__(self, other):
        return str(self) >= str(other)

    def __contains__(self, other):
        return str(other) in str(self)
